
#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


from tests import config
from firebase import initialize_app


def initiate_app_with_service_account_file():
	return initialize_app(config.SERVICE_CONFIG_WITH_FILE_PATH)


def make_auth(service_account=False):
	if service_account:
		c = config.SERVICE_CONFIG
	else:
		c = config.SIMPLE_CONFIG

	return initialize_app(c).auth()


def make_db(service_account=False):
	if service_account:
		c = config.SERVICE_CONFIG
	else:
		c = config.SIMPLE_CONFIG

	return initialize_app(c).database()


def make_ds(service_account=False):
	if service_account:
		c = config.SERVICE_CONFIG
	else:
		c = config.SIMPLE_CONFIG

	return initialize_app(c).firestore()


def make_storage(service_account=False):
	if service_account:
		c = config.SERVICE_CONFIG
	else:
		c = config.SIMPLE_CONFIG

	return initialize_app(c).storage()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import toml

sys.path.insert(0, os.path.abspath('..'))

project_config = toml.load('..//pyproject.toml')

# -- Project information -----------------------------------------------------

project = 'firebase-rest-api'
copyright = '2022, Asif Arman Rahman'
author = 'Asif Arman Rahman'

# The full version, including alpha/beta/rc tags
release = project_config['project']['version']


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
	'sphinx.ext.autodoc',
	'sphinx.ext.autosectionlabel',
	'sphinx.ext.intersphinx',
	'sphinx.ext.napoleon',
	'sphinx.ext.viewcode',
	'sphinx_design',
]

# auto section label configuration
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 3

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
	'google': ("https://googleapis.dev/python/google-auth/latest/", None),
	'google.cloud': ("https://googleapis.dev/python/storage/latest/", None),
	'google.cloud.firestore': ("https://googleapis.dev/python/firestore/latest/", None),
	'python': ("https://docs.python.org/3/", None),
	'requests': ("https://requests.readthedocs.io/en/stable/", None),
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# custom static css files to add
html_css_files = [
	"https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.1.1/css/fontawesome.min.css",
	"css/my-styles.css",
]



#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------
import os.path

import pytest


class TestStorage:

	test_user = None

	def test_user_for_storage(self, auth):
		self.__class__.test_user = auth.sign_in_anonymous()
		assert self.__class__.test_user is not None

	def test_child(self, storage):
		assert storage.child('firebase-test-001')

	def test_put(self, storage):
		assert storage.child('uploaded-file.txt').put("tests/static/test-file.txt", self.__class__.test_user.get('idToken'))

	def test_get_url(self, storage):
		assert storage.child('firebase-test-001').child('uploaded-file.txt').get_url(self.__class__.test_user.get('idToken'))

	def test_download(self, storage):
		assert storage.child('firebase-test-001').child('uploaded-file.txt').download('tests/static/downloaded.txt', self.__class__.test_user.get('idToken')) is None
		assert os.path.exists('tests/static/downloaded.txt')

	def test_delete(self, storage):
		os.remove('tests/static/downloaded.txt')
		assert storage.child('firebase-test-001/uploaded-file.txt').delete(self.__class__.test_user.get('idToken')) is None

	def test_list_of_files(self, storage):
		with pytest.raises(AttributeError) as exc_info:
			storage.list_files()
		assert 'bucket' in str(exc_info.value)

	def test_clean_user(self, auth):
		assert auth.delete_user_account(self.__class__.test_user.get('idToken'))


class TestStorageAdmin:

	def test_child(self, storage_admin):
		assert storage_admin.child('firebase-test-001')

	def test_put(self, storage_admin):
		assert storage_admin.child('uploaded-file.txt').put("tests/static/test-file.txt") is None

	def test_get_url(self, storage_admin):
		assert storage_admin.child('firebase-test-001').child('uploaded-file.txt').get_url(None)

	def test_download(self, storage_admin):
		assert storage_admin.child('firebase-test-001').child('uploaded-file.txt').download('tests/static/downloaded.txt') is None
		assert os.path.exists('tests/static/downloaded.txt')

	def test_delete(self, storage_admin):
		os.remove('tests/static/downloaded.txt')
		assert storage_admin.child('firebase-test-001/uploaded-file.txt').delete() is None

	def test_list_of_files(self, storage_admin):
		assert storage_admin.list_files()



#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


import pytest

from tests.tools import make_auth, make_db, make_ds, make_storage
from tests.config import TEST_USER_EMAIL, TEST_USER_PASSWORD


@pytest.fixture(scope='session')
def auth():
	return make_auth()


@pytest.fixture(scope='session')
def auth_admin():
	return make_auth(True)


@pytest.fixture(scope='session')
def db():
	# To make it easier to test, we keep the test restricted to firebase_tests
	# Because of the current mutations on calls, we return it as a function.
	try:
		yield lambda: make_db(service_account=True).child('firebase_tests')
	finally:
		make_db(service_account=True).child('firebase_tests').remove()


@pytest.fixture(scope='session')
def email():
	return TEST_USER_EMAIL


@pytest.fixture(scope='session')
def ds():
	return make_ds()


@pytest.fixture(scope='session')
def ds_admin():
	return make_ds(True)


@pytest.fixture(scope='session')
def password():
	return TEST_USER_PASSWORD


@pytest.fixture(scope='session')
def storage():
	return make_storage()


@pytest.fixture(scope='session')
def storage_admin():
	return make_storage(service_account=True)




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


"""
A simple python wrapper for Google's `Firebase Database REST API`_

.. _Firebase Database REST API: 
	https://firebase.google.com/docs/reference/rest/database
"""

import math
import json
import time
from random import randrange
from urllib.parse import urlencode
from google.auth.transport.requests import Request

from ._stream import Stream
from ._db_convert import FirebaseResponse
from firebase._exception import raise_detailed_error
from ._db_convert import convert_to_firebase, convert_list_to_firebase


class Database:
	""" Firebase Database Service


	:type credentials: :class:`~google.oauth2.service_account.Credentials`
	:param credentials: Service Account Credentials.

	:type database_url: str
	:param database_url: ``databaseURL`` from Firebase configuration.

	:type requests: :class:`~requests.Session`
	:param requests: Session to make HTTP requests.
	"""

	def __init__(self, credentials, database_url, requests):
		""" Constructor """

		if not database_url.endswith('/'):
			url = ''.join([database_url, '/'])
		else:
			url = database_url

		self.credentials = credentials
		self.database_url = url
		self.requests = requests

		self.path = ""
		self.build_query = {}
		self.last_push_time = 0
		self.last_rand_chars = []

	def order_by_key(self):
		""" Filter data by their keys.

		| For more details:
		| |filtering_by_key|_

		.. |filtering_by_key| replace:: 
			Firebase Documentation | Retrieve Data | Filtering
			Data | filtering_by_key

		.. _filtering_by_key: 
			https://firebase.google.com/docs/database/rest/retrieve-data#filtering-by-key


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["orderBy"] = "$key"

		return self

	def order_by_value(self):
		""" Filter data by the value of their child keys.

		| For more details:
		| |filtering-by-value|_
		
		.. |filtering-by-value| replace::
			Firebase Documentation | Retrieve Data | Filtering
			Data | filtering-by-value

		.. _filtering-by-value: 
			https://firebase.google.com/docs/database/rest/retrieve-data#filtering-by-value


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["orderBy"] = "$value"

		return self

	def order_by_child(self, order):
		""" Filter data by a common child key.

		| For more details:
		| |filtering-by-a-specified-child-key|_

		.. |filtering-by-a-specified-child-key| replace::
			Firebase Documentation | Retrieve Data | Filtering
			Data | filtering-by-a-specified-child-key

		.. _filtering-by-a-specified-child-key: 
			https://firebase.google.com/docs/database/rest/retrieve-data#filtering-by-a-specified-child-key


		:type order: str
		:param order: Child key name.


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["orderBy"] = order

		return self

	def start_at(self, start):
		""" Filter data where child key value starts from specified
		value.

		| For more details:
		| |range-queries|_

		.. |range-queries| replace::
			Firebase Documentation | Retrieve Data | Complex
			Filtering | range-queries

		.. _range-queries: 
			https://firebase.google.com/docs/database/rest/retrieve-data#range-queries


		:type start: int or float or str
		:param start: Arbitrary starting points for queries.


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["startAt"] = start

		return self

	def end_at(self, end):
		""" Filter data where child key value ends at specified
		value.

		| For more details:
		| |range-queries|_


		:type end: int or float or str
		:param end: Arbitrary ending points for queries.


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["endAt"] = end

		return self

	def equal_to(self, equal):
		""" Filter data where child key value is equal to specified
		value.

		| For more details:
		| |range-queries|_


		:type equal: int or float or str
		:param equal: Arbitrary point for queries.


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["equalTo"] = equal

		return self

	def limit_to_first(self, limit_first):
		""" Filter the number of data to receive from top.

		| For more details:
		| |limit-queries|_

		.. |limit-queries| replace::
			Firebase Documentation | Retrieve Data | Complex
			Filtering | limit-queries

		.. _limit-queries: 
			https://firebase.google.com/docs/database/rest/retrieve-data#limit-queries


		:type limit_first: int
		:param limit_first: Maximum number of children to select
			from top.


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["limitToFirst"] = limit_first

		return self

	def limit_to_last(self, limit_last):
		""" Filter the number of data to receive from bottom.

		| For more details:
		| |limit-queries|_


		:type limit_last: int
		:param limit_last: Maximum number of children to select
			from bottom.


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["limitToLast"] = limit_last

		return self

	def shallow(self):
		""" Limit the depth of the response.

		| For more details:
		| |section-param-shallow|_

		.. |section-param-shallow| replace::
			Firebase Database REST API | Query Parameters |
			section-param-shallow |

		.. _section-param-shallow:
			https://firebase.google.com/docs/reference/rest/database#section-param-shallow


		:return: A reference to the instance object.
		:rtype: Database
		"""

		self.build_query["shallow"] = True

		return self

	def child(self, *args):
		""" Build paths to your data.


		:type args: str
		:param args: Positional arguments to build path to database.


		:return: A reference to the instance object.
		:rtype: Database
		"""

		new_path = "/".join([str(arg) for arg in args])

		if self.path:
			self.path += "/{}".format(new_path)

		else:
			if new_path.startswith("/"):
				new_path = new_path[1:]

			self.path = new_path

		return self

	def build_request_url(self, token):
		""" Builds Request URL for query.


		:type token: str
		:param token: Firebase Auth User ID Token


		:return: Request URL
		:rtype: str
		"""

		parameters = {}

		if token:
			parameters['auth'] = token

		for param in list(self.build_query):
			if type(self.build_query[param]) is str:
				parameters[param] = '"' + self.build_query[param] + '"'

			elif type(self.build_query[param]) is bool:
				parameters[param] = "true" if self.build_query[param] else "false"

			else:
				parameters[param] = self.build_query[param]

		# reset path and build_query for next query
		request_ref = '{0}{1}.json?{2}'.format(self.database_url, self.path, urlencode(parameters))

		self.path = ""
		self.build_query = {}

		return request_ref

	def build_headers(self, token=None):
		""" Build Request Header.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.


		:return: Request Header.
		:rtype: dict
		"""

		headers = {"content-type": "application/json; charset=UTF-8"}

		if not token and self.credentials:

			if not self.credentials.valid:
				self.credentials.refresh(Request())

			access_token = self.credentials.token
			headers['Authorization'] = 'Bearer ' + access_token

		return headers

	def get(self, token=None, json_kwargs={}):
		""" Read data from database.

		| For more details:
		| |section-get|_

		.. |section-get| replace::
			Firebase Database REST API | GET - Reading Data

		.. _section-get:
			https://firebase.google.com/docs/reference/rest/database#section-get


		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.

		:type json_kwargs: dict
		:param json_kwargs: (Optional) Keyword arguments to send to 
			:func:`json.dumps` method for deserialization of data, 
			defaults to :data:`{}` (empty :class:`dict` object).
				

		:return: The data associated with the path.
		:rtype: dict
		"""

		build_query = self.build_query
		query_key = self.path.split("/")[-1]
		request_ref = self.build_request_url(token)

		# headers
		headers = self.build_headers(token)

		# do request
		request_object = self.requests.get(request_ref, headers=headers)

		raise_detailed_error(request_object)
		request_dict = request_object.json(**json_kwargs)

		# if primitive or simple query return
		if isinstance(request_dict, list):
			return FirebaseResponse(convert_list_to_firebase(request_dict), query_key)

		if not isinstance(request_dict, dict):
			return FirebaseResponse(request_dict, query_key)

		if not build_query:
			return FirebaseResponse(convert_to_firebase(request_dict.items()), query_key)

		# return keys if shallow
		if build_query.get("shallow"):
			return FirebaseResponse(request_dict.keys(), query_key)

		# otherwise sort
		sorted_response = None

		if build_query.get("orderBy"):
			if build_query["orderBy"] == "$key":
				sorted_response = sorted(request_dict.items(), key=lambda item: item[0])
			elif build_query["orderBy"] == "$value":
				sorted_response = sorted(request_dict.items(), key=lambda item: item[1])
			else:
				sorted_response = sorted(request_dict.items(), key=lambda item: (build_query["orderBy"] in item[1], item[1].get(build_query["orderBy"], "")))

		return FirebaseResponse(convert_to_firebase(sorted_response), query_key)

	def push(self, data, token=None, json_kwargs={}):
		""" Add data to database.

		This method adds a Firebase Push ID at the end of the specified 
		path, and then adds/stores the data in database, unlike 
		:meth:`set` which does not use a Firebase Push ID.

		| For more details:
		| |section-post|_

		.. |section-post| replace::
			Firebase Database REST API | POST - Pushing Data

		.. _section-post:
			https://firebase.google.com/docs/reference/rest/database#section-post


		:type data: dict
		:param data: Data to be stored in database.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.

		:type json_kwargs: dict
		:param json_kwargs: (Optional) Keyword arguments to send to 
			:func:`json.dumps` method for serialization of data, 
			defaults to :data:`{}` (empty :class:`dict` object).


		:return: Child key (Firebase Push ID) name of the data.
		:rtype: dict
		"""

		request_ref = self.check_token(self.database_url, self.path, token)

		self.path = ""

		headers = self.build_headers(token)
		request_object = self.requests.post(request_ref, headers=headers, data=json.dumps(data, **json_kwargs).encode("utf-8"))

		raise_detailed_error(request_object)

		return request_object.json()

	def set(self, data, token=None, json_kwargs={}):
		""" Add data to database.

		This method writes the data in database in the specified 
		path, unlike :meth:`push` which creates a Firebase Push ID then 
		writes the data to database.

		| For more details:
		| |section-put|_

		.. |section-put| replace::
			Firebase Database REST API | PUT - Writing Data

		.. _section-put:
			https://firebase.google.com/docs/reference/rest/database#section-put


		:type data: dict
		:param data: Data to be stored in database.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.

		:type json_kwargs: dict
		:param json_kwargs: (Optional) Keyword arguments to send to 
			:func:`json.dumps` method for serialization of data, 
			defaults to :data:`{}` (empty :class:`dict` object).


		:return: Successful attempt returns the ``data`` specified to 
			add to database.
		:rtype: dict
		"""

		request_ref = self.check_token(self.database_url, self.path, token)

		self.path = ""

		headers = self.build_headers(token)
		request_object = self.requests.put(request_ref, headers=headers, data=json.dumps(data, **json_kwargs).encode("utf-8"))

		raise_detailed_error(request_object)

		return request_object.json()

	def update(self, data, token=None, json_kwargs={}):
		""" Update stored data of database.

		| For more details:
		| |section-patch|_

		.. |section-patch| replace:: 
			Firebase Database REST API | PATCH - Updating Data

		.. _section-patch: 
			https://firebase.google.com/docs/reference/rest/database#section-patch


		:type data: dict
		:param data: Data to be updated.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.

		:type json_kwargs: dict
		:param json_kwargs: (Optional) Keyword arguments to send to 
			:func:`json.dumps` method for serialization of data, 
			defaults to :data:`{}` (empty :class:`dict` object).


		:return: Successful attempt returns the data specified to 
			update.
		:rtype: dict
		"""

		request_ref = self.check_token(self.database_url, self.path, token)

		self.path = ""

		headers = self.build_headers(token)
		request_object = self.requests.patch(request_ref, headers=headers, data=json.dumps(data, **json_kwargs).encode("utf-8"))

		raise_detailed_error(request_object)

		return request_object.json()

	def remove(self, token=None):
		""" Delete data from database.

		| For more details:
		| |section-delete|_

		.. |section-delete| replace::
			Firebase Database REST API | DELETE - Removing Data

		.. _section-delete: 
			https://firebase.google.com/docs/reference/rest/database#section-delete


		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.


		:return: Successful attempt returns :data:`None`.
		:rtype: :data:`None`
		"""

		request_ref = self.check_token(self.database_url, self.path, token)

		self.path = ""

		headers = self.build_headers(token)
		request_object = self.requests.delete(request_ref, headers=headers)

		raise_detailed_error(request_object)

		return request_object.json()

	def stream(self, stream_handler, token=None, stream_id=None, is_async=True):
		request_ref = self.build_request_url(token)

		return Stream(request_ref, stream_handler, self.build_headers, stream_id, is_async)

	def check_token(self, database_url, path, token):
		""" Builds Request URL to write/update/remove data.


		:type database_url: str
		:param database_url: ``databaseURL`` from Firebase 
			configuration.

		:type path: str
		:param path: Path to data.

		:type token: str
		:param token: Firebase Auth User ID Token


		:return: Request URL
		:rtype: str
		"""

		if token:
			return '{0}{1}.json?auth={2}'.format(database_url, path, token)
		else:
			return '{0}{1}.json'.format(database_url, path)

	def generate_key(self):
		""" Generate Firebase's push IDs.

		| For more details:
		| |firebase-push-id|_

		.. |firebase-push-id| replace::
			Firebase Blog | The 2^120 Ways to Ensure Unique Identifiers

		.. _firebase-push-id:
			https://firebase.blog/posts/2015/02/the-2120-ways-to-ensure-unique_68


		:return: Firebase's push IDs
		:rtype: str
		"""

		push_chars = '-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz'

		now = int(time.time() * 1000)
		duplicate_time = now == self.last_push_time

		self.last_push_time = now
		time_stamp_chars = [0] * 8

		for i in reversed(range(0, 8)):
			time_stamp_chars[i] = push_chars[now % 64]
			now = int(math.floor(now / 64))

		new_id = "".join(time_stamp_chars)

		if not duplicate_time:
			self.last_rand_chars = [randrange(64) for _ in range(12)]
		else:
			for i in range(0, 11):

				if self.last_rand_chars[i] == 63:
					self.last_rand_chars[i] = 0

				self.last_rand_chars[i] += 1

		for i in range(0, 12):
			new_id += push_chars[self.last_rand_chars[i]]

		return new_id

	def sort(self, origin, by_key, reverse=False):
		""" Further sort data based on a child key value.


		:type origin: dict
		:param origin: Data to be sorted (generally the output from 
			:meth:`get` method).

		:type by_key: str
		:param by_key: Child key name to sort by.

		:type reverse: bool
		:param reverse: (Optional) Whether to return data in descending 
			order, defaults to :data:`False` (data is returned in 
			ascending order).


		:return: Sorted version of the data.
		:rtype: dict
		"""

		# unpack firebase objects
		firebases = origin.each()

		new_list = []

		for firebase in firebases:
			new_list.append(firebase.item)

		# sort
		data = sorted(dict(new_list).items(), key=lambda item: item[1][by_key], reverse=reverse)

		return FirebaseResponse(convert_to_firebase(data), origin.key())

	def get_etag(self, token=None):
		""" Fetches Firebase ETag at a specified location.

		| For more details:
		| |section-cond-etag|_

		.. |section-cond-etag| replace::
			Firebase Database REST API | Conditional Requests | 
			#section-cond-etag

		.. _section-cond-etag: 
			https://firebase.google.com/docs/reference/rest/database#section-cond-etag


		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.


		:return: Firebase ETag
		:rtype: str
		"""

		request_ref = self.build_request_url(token)

		headers = self.build_headers(token)
		# extra header to get ETag
		headers['X-Firebase-ETag'] = 'true'
		request_object = self.requests.get(request_ref, headers=headers)

		raise_detailed_error(request_object)

		return request_object.headers['ETag']

	def conditional_set(self, data, etag, token=None, json_kwargs={}):
		""" Conditionally add data to database.

		| For more details:
		| |section-expected-responses|_

		.. |section-expected-responses| replace::
			Firebase Database REST API | Conditional Requests | 
			section-expected-responses

		.. _section-expected-responses:
			https://firebase.google.com/docs/reference/rest/database#section-expected-responses


		:type data: dict
		:param data: Data to be stored in database.
		
		:type etag: str
		:param etag: Unique identifier for specific data at a
			specified location.
		
		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.
		
		:type json_kwargs: dict
		:param json_kwargs: (Optional) Keyword arguments to send to 
			:meth:`json.dumps` methods for serialization of data, 
			defaults to ``{}`` (empty :class:`dict` object).


		:return: Successful attempt returns the data specified to store,
			failed attempt (due to ETag mismatch) returns the current 
			``ETag`` for the specified path.
		:rtype: dict
		"""

		request_ref = self.check_token(self.database_url, self.path, token)

		self.path = ""

		headers = self.build_headers(token)
		headers['if-match'] = etag

		request_object = self.requests.put(request_ref, headers=headers, data=json.dumps(data, **json_kwargs).encode("utf-8"))

		# ETag didn't match, so we should return the correct one for the user to try again
		if request_object.status_code == 412:
			return {'ETag': request_object.headers['ETag']}

		raise_detailed_error(request_object)

		return request_object.json()

	def conditional_remove(self, etag, token=None):
		""" Conditionally delete data from database.

		| For more details:
		| |section-expected-responses|_


		:type etag: str
		:param etag: Unique identifier for specific data at a
			specified location.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.


		:return: Successful attempt returns :data:`None`, in case of ETag
			mismatch an updated ETag for the specific data is
			returned in :class:`dict` object
		:rtype: :data:`None`
		"""

		request_ref = self.check_token(self.database_url, self.path, token)

		self.path = ""

		headers = self.build_headers(token)
		headers['if-match'] = etag
		request_object = self.requests.delete(request_ref, headers=headers)

		# ETag didn't match, so we should return the correct one for the user to try again
		if request_object.status_code == 412:
			return {'ETag': request_object.headers['ETag']}

		raise_detailed_error(request_object)

		return request_object.json()




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


"""A simple python wrapper for Google's `Firebase`_ REST APIs.

.. _Firebase: https://firebase.google.com/
"""

from .auth import Auth
from .storage import Storage
from .database import Database
from .firestore import Firestore
from ._custom_requests import _custom_request
from ._service_account_credentials import _service_account_creds_from_secret


def initialize_app(config):
	"""Initializes and returns a new Firebase instance.

	:type config: dict
	:param config: Firebase configuration

	:return: A newly initialized instance of Firebase.
	:rtype: Firebase
	"""

	return Firebase(config)


class Firebase:
	""" Firebase Interface

	:type config: dict
	:param config: Firebase configuration
	"""

	def __init__(self, config):
		""" Constructor """

		self.api_key = config["apiKey"]
		self.auth_domain = config["authDomain"]
		self.database_url = config["databaseURL"]
		self.project_id = config["projectId"]
		self.storage_bucket = config["storageBucket"]

		self.credentials = None
		self.requests = _custom_request()

		if config.get("serviceAccount"):
			self.credentials = _service_account_creds_from_secret(config['serviceAccount'])

	def auth(self, client_secret=None):
		"""Initializes and returns a new Firebase Authentication
		instance.

		:type client_secret: str or dict
		:param client_secret: (Optional) File path to or the dict
			object from social client secret file, defaults to
			:data:`None`.


		:return: A newly initialized instance of Auth.
		:rtype: Auth
		"""

		return Auth(self.api_key, self.credentials, self.requests, client_secret=client_secret)

	def database(self):
		"""Initializes and returns a new Firebase Realtime Database
		instance.

		:return: A newly initialized instance of Database.
		:rtype: Database
		"""

		return Database(self.credentials, self.database_url, self.requests)

	def firestore(self):
		"""Initializes and returns a new Firebase Cloud Firestore
		instance.

		:return: A newly initialized instance of Firestore.
		:rtype: Firestore
		"""

		return Firestore(self.api_key, self.credentials, self.project_id, self.requests)

	def storage(self):
		"""Initializes and returns a new Firebase Storage instance.

		:return: A newly initialized instance of Storage.
		:rtype: Storage
		"""

		return Storage(self.credentials, self.requests, self.storage_bucket)




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


import json
import time
import threading

from ._keep_auth_session import KeepAuthSession
from ._closable_sse_client import ClosableSSEClient


class Stream:

	def __init__(self, url, stream_handler, build_headers, stream_id, is_async):
		self.build_headers = build_headers
		self.url = url
		self.stream_handler = stream_handler
		self.stream_id = stream_id
		self.sse = None
		self.thread = None

		if is_async:
			self.start()
		else:
			self.start_stream()

	def make_session(self):
		"""
		Return a custom session object to be passed to the ClosableSSEClient.
		"""
		session = KeepAuthSession()

		return session

	def start(self):
		self.thread = threading.Thread(target=self.start_stream)
		self.thread.start()

		return self

	def start_stream(self):
		self.sse = ClosableSSEClient(self.url, session=self.make_session(), build_headers=self.build_headers)

		for msg in self.sse:
			if msg:
				msg_data = json.loads(msg.data)
				msg_data["event"] = msg.event

				if self.stream_id:
					msg_data["stream_id"] = self.stream_id

				self.stream_handler(msg_data)

	def close(self):
		while not self.sse and not hasattr(self.sse, 'resp'):
			time.sleep(0.001)

		self.sse.running = False
		self.sse.close()
		self.thread.join()

		return self




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


import re
import six
import time
import requests
import warnings


# Technically, we should support streams that mix line endings.  This regex,
# however, assumes that a system will provide consistent line endings.
end_of_field = re.compile(r'\r\n\r\n|\r\r|\n\n')


class SSEClient(object):

	def __init__(self, url, session, build_headers, last_id=None, retry=3000, **kwargs):
		self.url = url
		self.last_id = last_id
		self.retry = retry
		self.running = True

		# Optional support for passing in a requests.Session()
		self.session = session

		# function for building auth header when token expires
		self.build_headers = build_headers
		self.start_time = None

		# Any extra kwargs will be fed into the requests.get call later.
		self.requests_kwargs = kwargs

		# The SSE spec requires making requests with Cache-Control: nocache
		if 'headers' not in self.requests_kwargs:
			self.requests_kwargs['headers'] = {}

		self.requests_kwargs['headers']['Cache-Control'] = 'no-cache'

		# The 'Accept' header is not required, but explicit > implicit
		self.requests_kwargs['headers']['Accept'] = 'text/event-stream'

		# Keep data here as it streams in
		self.buf = u''

		self._connect()

	def _connect(self):
		if self.last_id:
			self.requests_kwargs['headers']['Last-Event-ID'] = self.last_id

		headers = self.build_headers()
		self.requests_kwargs['headers'].update(headers)

		# Use session if set.  Otherwise, fall back to `requests` module.
		self.requester = self.session or requests
		self.resp = self.requester.get(self.url, stream=True, **self.requests_kwargs)

		self.resp_iterator = self.resp.iter_content(decode_unicode=True)

		# TODO: Ensure we're handling redirects.  Might also stick the 'origin'
		# attribute on Events like the Javascript spec requires.
		self.resp.raise_for_status()

	def _event_complete(self):
		return re.search(end_of_field, self.buf) is not None

	def __iter__(self):
		return self

	def __next__(self):
		while not self._event_complete():

			try:
				nextchar = next(self.resp_iterator)
				self.buf += nextchar
			except (StopIteration, requests.RequestException):
				time.sleep(self.retry / 1000.0)
				self._connect()

				# The SSE spec only supports resuming from a whole message, so
				# if we have half a message we should throw it out.
				head, sep, tail = self.buf.rpartition('\n')
				self.buf = head + sep
				continue

		split = re.split(end_of_field, self.buf)
		head = split[0]
		tail = "".join(split[1:])

		self.buf = tail
		msg = Event.parse(head)

		if msg.data == "credential is no longer valid":
			self._connect()
			return None

		if msg.data == 'null':
			return None

		# If the server requests a specific retry delay, we need to honor it.
		if msg.retry:
			self.retry = msg.retry

		# last_id should only be set if included in the message.  It's not
		# forgotten if a message omits it.
		if msg.id:
			self.last_id = msg.id

		return msg

	if six.PY2:
		next = __next__


class Event(object):

	sse_line_pattern = re.compile('(?P<name>[^:]*):?( ?(?P<value>.*))?')

	def __init__(self, data='', event='message', id=None, retry=None):
		self.data = data
		self.event = event
		self.id = id
		self.retry = retry

	def dump(self):
		lines = []

		if self.id:
			lines.append('id: %s' % self.id)

		# Only include an event line if it's not the default already.
		if self.event != 'message':
			lines.append('event: %s' % self.event)

		if self.retry:
			lines.append('retry: %s' % self.retry)

		lines.extend('data: %s' % d for d in self.data.split('\n'))

		return '\n'.join(lines) + '\n\n'

	@classmethod
	def parse(cls, raw):
		"""
		Given a possibly-multiline string representing an SSE message, parse it
		and return a `Event` object.
		"""
		msg = cls()

		for line in raw.split('\n'):
			m = cls.sse_line_pattern.match(line)

			if m is None:
				# Malformed line.  Discard but warn.
				warnings.warn('Invalid SSE line: "%s"' % line, SyntaxWarning)
				continue

			name = m.groupdict()['name']
			value = m.groupdict()['value']

			if name == '':
				# line began with a ":", so is a comment.  Ignore
				continue

			if name == 'data':
				# If we already have some data, then join to it with a newline.
				# Else this is it.
				if msg.data:
					msg.data = '%s\n%s' % (msg.data, value)
				else:
					msg.data = value

			elif name == 'event':
				msg.event = value

			elif name == 'id':
				msg.id = value

			elif name == 'retry':
				msg.retry = int(value)

		return msg

	def __str__(self):
		return self.data




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


from requests import Session


class KeepAuthSession(Session):
	"""
	A session that doesn't drop Authentication on redirects between domains.
	"""

	def rebuild_auth(self, prepared_request, response):
		pass




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


from decouple import config


# get this from firebase console
# go to project settings, general tab and click "Add Firebase to your web app"
SIMPLE_CONFIG = {
	"apiKey": config('FIREBASE_API_KEY'),
	"authDomain": config('FIREBASE_AUTH_DOMAIN'),
	"databaseURL": config('FIREBASE_DATABASE_URL'),
	"projectId": config('FIREBASE_PROJECT_ID'),
	"storageBucket": config('FIREBASE_STORAGE_BUCKET'),
}

# get this in json file from firebase console
# go to project settings, service accounts tab and click generate new private key
SERVICE_ACCOUNT = {
	"type": 'service_account',
	"project_id": config('FIREBASE_SERVICE_ACCOUNT_PROJECT_ID'),
	"private_key_id": config('FIREBASE_SERVICE_ACCOUNT_PRIVATE_KEY_ID'),
	"private_key": config('FIREBASE_SERVICE_ACCOUNT_PRIVATE_KEY').replace('\\n', '\n'),
	"client_email": config('FIREBASE_SERVICE_ACCOUNT_CLIENT_EMAIL'),
	"client_id": config('FIREBASE_SERVICE_ACCOUNT_CLIENT_ID'),
	"auth_uri": 'https://accounts.google.com/o/oauth2/auth',
	"token_uri": 'https://oauth2.googleapis.com/token',
	"auth_provider_x509_cert_url": 'https://www.googleapis.com/oauth2/v1/certs',
	"client_x509_cert_url": config('FIREBASE_SERVICE_ACCOUNT_CLIENT_X509_CERT_URL'),
}


SERVICE_CONFIG = dict(SIMPLE_CONFIG, serviceAccount=SERVICE_ACCOUNT)


# get this json file from firebase console
# go to project settings, service accounts tab and click generate new private key
SERVICE_ACCOUNT_PATH = "firebase-adminsdk.json"

SERVICE_CONFIG_WITH_FILE_PATH = dict(SIMPLE_CONFIG, serviceAccount=SERVICE_ACCOUNT_PATH)


TEST_USER_EMAIL = config('TEST_USER_EMAIL')
TEST_USER_PASSWORD = config('TEST_USER_PASSWORD')




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


import time
import random
import pytest
import datetime
from contextlib import contextmanager


@pytest.fixture(scope='function')
def db_sa(db):
	# To make it easier to test, we keep the test restricted to firebase_tests
	# Because of the current mutations on calls, we return it as a function.
	name = 'test_%05d' % random.randint(0, 99999)
	yield lambda: db().child(name)


@contextmanager
def make_stream(db, cbk):
	s = db.stream(cbk)
	try:
		yield s
	finally:
		s.close()


@contextmanager
def make_append_stream(db):
	l = []

	def cbk(event):
		l.append(event)

	with make_stream(db, cbk) as s:
		yield s, l


class TestSimpleGetAndPut:
	def test_simple_get(self, db_sa):
		assert db_sa().get().val() is None

	def test_put_succeed(self, db_sa):
		assert db_sa().set(True)

	def test_put_then_get_keeps_value(self, db_sa):
		db_sa().set("some_value")
		assert db_sa().get().val() == "some_value"

	def test_put_dictionary(self, db_sa):
		v = dict(a=1, b="2", c=dict(x=3.1, y="3.2"))
		db_sa().set(v)

		assert db_sa().get().val() == v

	@pytest.mark.skip
	def test_put_deeper_dictionary(self, db_sa):
		v = {'1': {'11': {'111': 42}}}
		db_sa().set(v)

		# gives: assert [None, {'11': {'111': 42}}] == {'1': {'11': {'111': 42}}}
		assert db_sa().get().val() == v


class TestJsonKwargs:

	def encoder(self, obj):
		if isinstance(obj, datetime.datetime):
			return {
				'__type__': obj.__class__.__name__,
				'value': obj.timestamp(),
			}
		return obj

	def decoder(self, obj):
		if '__type__' in obj and obj['__type__'] == datetime.datetime.__name__:
			return datetime.datetime.fromtimestamp(obj['value'])
		return obj

	def test_put_fail(self, db_sa):
		v = {'some_datetime': datetime.datetime.now()}
		with pytest.raises(TypeError):
			db_sa().set(v)

	def test_put_succeed(self, db_sa):
		v = {'some_datetime': datetime.datetime.now()}
		assert db_sa().set(v, json_kwargs={'default': str})

	def test_put_then_get_succeed(self, db_sa):
		v = {'another_datetime': datetime.datetime.now()}
		db_sa().set(v, json_kwargs={'default': self.encoder})
		assert db_sa().get(json_kwargs={'object_hook': self.decoder}).val() == v


class TestChildNavigation:
	def test_get_child_none(self, db_sa):
		assert db_sa().child('lorem').get().val() is None

	def test_get_child_after_pushing_data(self, db_sa):
		db_sa().set({'lorem': "a", 'ipsum': 2})

		assert db_sa().child('lorem').get().val() == "a"
		assert db_sa().child('ipsum').get().val() == 2

	def test_update_child(self, db_sa):
		db_sa().child('child').update({'c1/c11': 1, 'c1/c12': 2, 'c2': 3})

		assert db_sa().child('child').child('c1').get().val() == {'c11': 1, 'c12': 2}
		assert db_sa().child('child').child('c2').get().val() == 3

	def test_path_equivalence(self, db_sa):
		db_sa().set({'1': {'11': {'111': 42}}})

		assert db_sa().child('1').child('11').child('111').get().val() == 42
		assert db_sa().child('1/11/111').get().val() == 42
		assert db_sa().child('1', '11', '111').get().val() == 42
		assert db_sa().child(1, '11', '111').get().val() == 42


class TestStreaming:
	def test_create_stream_succeed(self, db_sa):
		with make_append_stream(db_sa()) as (stream, l):
			time.sleep(2)
			assert stream is not None

	def test_does_initial_call(self, db_sa):
		with make_append_stream(db_sa()) as (stream, l):
			time.sleep(2)
			assert len(l) == 1

	def test_responds_to_update_calls(self, db_sa):
		with make_append_stream(db_sa()) as (stream, l):
			testdata = {"1": "a", "1_2": "b"}
			db_sa().set(testdata)
			db_sa().update({"2": "c"})
			db_sa().push("3")

			time.sleep(2)

			assert len(l) > 0

			# a race condition from an earlier test causes an object with no data put to be first in the list sometimes
			if l[0]["data"] == None:
				assert len(l) == 4
				assert l[1]["data"] == testdata
			else:
				assert len(l) == 3
				assert l[0]["data"] == testdata


class TestConditionalRequest:
	def test_conditional_set_succeed(self, db_sa):
		etag = db_sa().get_etag()
		result = db_sa().conditional_set({'1': 'a'}, etag)

		assert db_sa().child('1').get().val() == 'a'

	def test_conditional_set_fail(self, db_sa):
		etag = '{}123'.format(db_sa().get_etag())
		result = db_sa().conditional_set({'2': 'b'}, etag)

		assert 'ETag' in result

	def test_conditional_remove_succeed(self, db_sa):
		etag = db_sa().child('1').get_etag()
		result = db_sa().child('1').conditional_remove(etag)

		assert db_sa().child('1').get().val() is None

	def test_conditional_remove_fail(self, db_sa):
		etag = '{}123'.format(db_sa().get_etag())
		result = db_sa().conditional_remove(etag)

		assert 'ETag' in result




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


import pytest
import requests.exceptions


class TestAuth:

	user = None
	anonymous_user = None

	def test_sign_in_with_non_existing_account_email_and_password(self, auth, email, password):
		with pytest.raises(requests.exceptions.HTTPError) as exc_info:
			auth.sign_in_with_email_and_password(email, password)
		assert "EMAIL_NOT_FOUND" in str(exc_info.value)

	def test_create_user_with_email_and_password(self, auth, email, password):
		assert auth.create_user_with_email_and_password(email, password)

	def test_create_user_with_existing_email_and_password(self, auth, email, password):
		with pytest.raises(requests.exceptions.HTTPError) as exc_info:
			auth.create_user_with_email_and_password(email, password)
		assert "EMAIL_EXISTS" in str(exc_info.value)

	def test_sign_in_with_email_and_wrong_password(self, auth, email):
		with pytest.raises(requests.exceptions.HTTPError) as exc_info:
			auth.sign_in_with_email_and_password(email, 'WrongPassword123')
		assert "INVALID_PASSWORD" in str(exc_info.value)

	def test_sign_in_with_email_and_password(self, auth, email, password):
		user = auth.sign_in_with_email_and_password(email, password)
		self.__class__.user = user
		assert user

	def test_sign_in_anonymous(self, auth):
		user = auth.sign_in_anonymous()
		self.__class__.anonymous_user = user
		assert user

	def test_create_custom_token(self, auth):
		with pytest.raises(AttributeError):
			auth.create_custom_token('CreateCustomToken1')

	def test_create_custom_token_with_claims(self, auth):
		with pytest.raises(AttributeError):
			auth.create_custom_token('CreateCustomToken2', {'premium': True})

	def test_sign_in_with_custom_token(self, auth):
		with pytest.raises(requests.exceptions.HTTPError):
			auth.sign_in_with_custom_token(None)

	def test_refresh(self, auth):
		assert auth.refresh(self.__class__.user.get('refreshToken'))

	def test_get_account_info(self, auth):
		assert auth.get_account_info(self.__class__.user.get('idToken'))

	def test_send_email_verification(self, auth):
		assert auth.send_email_verification(self.__class__.user.get('idToken'))

	def test_send_password_reset_email(self, auth):
		assert auth.send_password_reset_email(self.__class__.user.get('email'))

	@pytest.mark.xfail
	def test_verify_password_reset_code(self, auth):
		assert auth.verify_password_reset_code('123456', 'NewTestPassword123')

	def test_update_profile_display_name(self, auth):
		new_name = 'Test User'
		user = auth.update_profile(self.__class__.user.get('idToken'), display_name=new_name)
		assert user
		assert new_name == user['displayName']

	def test_delete_user_account(self, auth):
		assert auth.delete_user_account(self.__class__.user.get('idToken'))
		assert auth.delete_user_account(self.__class__.anonymous_user.get('idToken'))


class TestAuthAdmin:

	user = None
	anonymous_user = None
	custom_token = None
	custom_token_with_claims = None
	custom_user = None
	custom_user_with_claims = None

	def test_sign_in_with_non_existing_account_email_and_password(self, auth_admin, email, password):
		with pytest.raises(requests.exceptions.HTTPError) as exc_info:
			auth_admin.sign_in_with_email_and_password(email, password)
		assert "EMAIL_NOT_FOUND" in str(exc_info.value)

	def test_create_user_with_email_and_password(self, auth_admin, email, password):
		assert auth_admin.create_user_with_email_and_password(email, password)

	def test_create_user_with_existing_email_and_password(self, auth_admin, email, password):
		with pytest.raises(requests.exceptions.HTTPError) as exc_info:
			auth_admin.create_user_with_email_and_password(email, password)
		assert "EMAIL_EXISTS" in str(exc_info.value)

	def test_sign_in_with_email_and_wrong_password(self, auth_admin, email):
		with pytest.raises(requests.exceptions.HTTPError) as exc_info:
			auth_admin.sign_in_with_email_and_password(email, 'WrongPassword123')
		assert "INVALID_PASSWORD" in str(exc_info.value)

	def test_sign_in_with_email_and_password(self, auth_admin, email, password):
		user = auth_admin.sign_in_with_email_and_password(email, password)
		self.__class__.user = user
		assert user

	def test_sign_in_anonymous(self, auth_admin):
		user = auth_admin.sign_in_anonymous()
		self.__class__.anonymous_user = user
		assert user

	def test_create_custom_token(self, auth_admin):
		token = auth_admin.create_custom_token('CreateCustomToken1')
		self.__class__.custom_token = token
		assert token

	def test_create_custom_token_with_claims(self, auth_admin):
		token = auth_admin.create_custom_token('CreateCustomToken2', {'premium': True})
		self.__class__.custom_token_with_claims = token
		assert token

	def test_sign_in_with_custom_token(self, auth_admin):
		user1 = auth_admin.sign_in_with_custom_token(self.__class__.custom_token)
		user2 = auth_admin.sign_in_with_custom_token(self.__class__.custom_token_with_claims)

		self.__class__.custom_user = user1
		self.__class__.custom_user_with_claims = user2

		assert user1
		assert user2

	def test_refresh(self, auth_admin):
		assert auth_admin.refresh(self.__class__.user.get('refreshToken'))

	def test_get_account_info(self, auth_admin):
		assert auth_admin.get_account_info(self.__class__.user.get('idToken'))

	def test_send_email_verification(self, auth_admin):
		assert auth_admin.send_email_verification(self.__class__.user.get('idToken'))

	def test_send_password_reset_email(self, auth_admin):
		assert auth_admin.send_password_reset_email(self.__class__.user.get('email'))

	@pytest.mark.xfail
	def test_verify_password_reset_code(self, auth_admin):
		assert auth_admin.verify_password_reset_code('123456', 'NewTestPassword123')

	def test_update_profile_display_name(self, auth_admin):
		new_name = 'Test User'
		user = auth_admin.update_profile(self.__class__.user.get('idToken'), display_name=new_name)
		assert user
		assert new_name == user['displayName']

	def test_delete_user_account(self, auth_admin):
		assert auth_admin.delete_user_account(self.__class__.user.get('idToken'))
		assert auth_admin.delete_user_account(self.__class__.anonymous_user.get('idToken'))
		assert auth_admin.delete_user_account(self.__class__.custom_user.get('idToken'))
		assert auth_admin.delete_user_account(self.__class__.custom_user_with_claims.get('idToken'))




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


from requests.exceptions import HTTPError


def raise_detailed_error(request_object):
	try:
		request_object.raise_for_status()
	except HTTPError as e:
		# raise detailed error message
		# TODO: Check if we get a { "error" : "Permission denied." } and handle automatically
		raise HTTPError(e, request_object.text)




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


import pytest

from tests.tools import initiate_app_with_service_account_file, make_auth, make_db, make_storage


def test_initiate_app_with_service_account_file():
	with pytest.raises(FileNotFoundError) as exc_info:
		initiate_app_with_service_account_file()
	assert "No such file or directory: 'firebase-adminsdk.json'" in str(exc_info.value)


def test_setup_auth():
	auth = make_auth()
	user = auth.sign_in_anonymous()

	assert auth.delete_user_account(user['idToken'])


def test_setup_auth_admin():
	auth = make_auth(True)
	user = auth.sign_in_anonymous()

	assert auth.delete_user_account(user['idToken'])


def test_setup_db():
	db = make_db(True)

	assert db.get()


def test_setup_storage():
	storage = make_storage()

	with pytest.raises(AttributeError) as exc_info:
		storage.list_files()
	assert 'bucket' in str(exc_info.value)


def test_setup_storage_admin():
	storage = make_storage(True)

	assert storage.list_files()




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


class TestFirestoreAdmin:
	movies1 = {
		'name': 'Iron Man',
		'lead': {'name': 'Robert Downey Jr.'},
		'released': False,
		'year': 2008,
		'rating': 7.9,
		'prequel': None,
		'cast': ['Jon Favreau', 'Gwyneth Paltrow', 'Jeff Bridges', b'J.A.R.V.I.S', 'Terrence Howard']
	}

	movies2 = {
		'name': 'Thor',
		'lead': {'name': 'Chris Hemsworth'},
		'released': False,
		'year': 2011,
		'rating': 7.0,
		'cast': ['Tom Hiddleston', 'Natalie Portman', 'Anthony Hopkins', 'Jeremy Renner', 'Stellan Skarsgård', 'Idris Elba', 'Kat Dennings']
	}

	auto_doc_id = None

	def test_manual_doc_set(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document('001').set(self.__class__.movies1) is None

	def test_auto_doc_add(self, ds_admin):
		doc_id = ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').add(self.__class__.movies2)
		assert doc_id

		self.__class__.auto_doc_id = doc_id

	def test_manual_doc_get(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document('001').get() == self.__class__.movies1
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document(self.__class__.auto_doc_id).get() == self.__class__.movies2

	def test_collection_get(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').get() == [{'001': self.__class__.movies1}, {self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_collection_get_start_after(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('rating').start_after({'rating': 7.4}).get() == [{'001': self.__class__.movies1}]
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('rating').start_after({'rating': 6.9}).get() == [{self.__class__.auto_doc_id: self.__class__.movies2}, {'001': self.__class__.movies1}]
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('rating').start_after({'rating': 8.5}).get() == []

	def test_collection_get_start_at(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('rating').start_at({'rating': 7.4}).get() == [{'001': self.__class__.movies1}]
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('rating').start_at({'rating': 8.0}).get() == []
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('rating').start_at({'rating': 7.0}).get() == [{self.__class__.auto_doc_id: self.__class__.movies2}, {'001': self.__class__.movies1}]		

	def test_collection_get_select(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').select(['lead.name', 'released']).get() == [{'001': {'lead': self.__class__.movies1['lead'], 'released': self.__class__.movies1['released']}}, {self.__class__.auto_doc_id: {'lead': self.__class__.movies2['lead'], 'released': self.__class__.movies2['released']}}]

	def test_collection_get_offset(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('year').offset(1).get() == [{self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_collection_get_limit_to_first(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('year').limit_to_first(1).get() == [{'001': self.__class__.movies1}]

	def test_collection_get_limit_to_last(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('year').limit_to_last(1).get() == [{'001': self.__class__.movies1}]

	def test_collection_get_end_at(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('year').end_at({'year': 2010}).get() == [{'001': self.__class__.movies1}]
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('year').end_at({'year': 2021}).get() == [{'001': self.__class__.movies1}, {self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_collection_get_end_before(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').order_by('year').end_before({'year': 2023}).get() == [{'001': self.__class__.movies1}, {self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_collection_get_where(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').where('lead.name', 'in',  ['Benedict Cumberbatch', 'Robert Downey Jr.']).get() == [{'001': self.__class__.movies1}]
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').where('rating', '<=', 8.0).order_by('rating', direction='DESCENDING').get() == [{'001': self.__class__.movies1}, {self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_manual_doc_update(self, ds_admin):
		update_data = {'released': True}

		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document('001').update(update_data) is None
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document('001').get(field_paths=['released']) == update_data

		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document(self.__class__.auto_doc_id).update(update_data) is None
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document(self.__class__.auto_doc_id).get(field_paths=['released']) == update_data

	def test_manual_doc_get_filtered(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document('001').get(field_paths=['name']) == {'name': self.__class__.movies1['name']}
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document(self.__class__.auto_doc_id).get(field_paths=['name']) == {'name': self.__class__.movies2['name']}

	def test_manual_doc_delete(self, ds_admin):
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document('001').delete() is None
		assert ds_admin.collection('Marvels').document('Movies').collection('PhaseOne').document(self.__class__.auto_doc_id).delete() is None


class TestFirestoreAuth:
	movies1 = {
		'name': 'Dr. Strange',
		'lead': {'name': 'Benedict Cumberbatch'},
		'released': False,
		'year': 2016,
		'rating': 7.5,
		'prequel': None,
		'cast': ['Tilda Swinton', 'Rachel McAdams', 'Mads Mikkelsen', 'Chiwetel Ejiofor', 'Benedict Wong']
	}

	movies2 = {
		'name': 'Black Panther',
		'lead': {'name': 'Chadwick Boseman'},
		'released': False,
		'year': 2018,
		'rating': 7.3,
		'prequel': None,
		'cast': ['Michael B. Jordan', 'Sebastian Stan', 'Letitia Wright', 'Martin Freeman', 'Winston Duke']
	}

	user = None
	auto_doc_id = None

	def test_create_test_user(self, auth):
		user = auth.sign_in_anonymous()
		self.__class__.user = user
		assert user
		assert user.get('idToken')

	def test_manual_doc_set(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document('014').set(self.__class__.movies1, token=self.__class__.user.get('idToken')) is None

	def test_auto_doc_add(self, ds):
		doc_id = ds.collection('Marvels').document('Movies').collection('PhaseThree').add(self.__class__.movies2, token=self.__class__.user.get('idToken'))
		assert doc_id

		self.__class__.auto_doc_id = doc_id

	def test_manual_doc_get(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document('014').get(token=self.__class__.user.get('idToken')) == self.__class__.movies1
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document(self.__class__.auto_doc_id).get(token=self.__class__.user.get('idToken')) == self.__class__.movies2

	def test_collection_get(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}, {self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_manual_doc_get_filtered(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document('014').get(field_paths=['name'], token=self.__class__.user.get('idToken')) == {'name': self.__class__.movies1['name']}
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document(self.__class__.auto_doc_id).get(field_paths=['name'], token=self.__class__.user.get('idToken')) == {'name': self.__class__.movies2['name']}

	def test_collection_get_start_after(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('rating').start_after({'rating': 7.4}).get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}]
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('rating').start_after({'rating': 7.2}).get(token=self.__class__.user.get('idToken')) == [{self.__class__.auto_doc_id: self.__class__.movies2}, {'014': self.__class__.movies1}]
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('rating').start_after({'rating': 8.5}).get(token=self.__class__.user.get('idToken')) == []

	def test_collection_get_start_at(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('rating').start_at({'rating': 7.4}).get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}]
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('rating').start_at({'rating': 8.0}).get(token=self.__class__.user.get('idToken')) == []
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('rating').start_at({'rating': 7.0}).get(token=self.__class__.user.get('idToken')) == [{self.__class__.auto_doc_id: self.__class__.movies2}, {'014': self.__class__.movies1}]		

	def test_collection_get_select(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').select(['lead.name', 'released']).get(token=self.__class__.user.get('idToken')) == [{'014': {'lead': self.__class__.movies1['lead'], 'released': self.__class__.movies1['released']}}, {self.__class__.auto_doc_id: {'lead': self.__class__.movies2['lead'], 'released': self.__class__.movies2['released']}}]

	def test_collection_get_offset(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('year').offset(1).get(token=self.__class__.user.get('idToken')) == [{self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_collection_get_limit_to_first(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('year').limit_to_first(1).get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}]

	def test_collection_get_limit_to_last(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('year').limit_to_last(1).get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}]

	def test_collection_get_end_at(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('year').end_at({'year': 2010}).get(token=self.__class__.user.get('idToken')) == []
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('year').end_at({'year': 2021}).get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}, {self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_collection_get_end_before(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').order_by('year').end_before({'year': 2023}).get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}, {self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_collection_get_where(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').where('lead.name', 'in',  ['Benedict Cumberbatch', 'Robert Downey Jr.']).get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}]
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').where('rating', '<=', 8.0).order_by('rating', direction='DESCENDING').get(token=self.__class__.user.get('idToken')) == [{'014': self.__class__.movies1}, {self.__class__.auto_doc_id: self.__class__.movies2}]

	def test_manual_doc_update(self, ds):
		update_data = {'released': True}

		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document('014').update(update_data, token=self.__class__.user.get('idToken')) is None
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document('014').get(field_paths=['released'], token=self.__class__.user.get('idToken')) == update_data

		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document(self.__class__.auto_doc_id).update(update_data, token=self.__class__.user.get('idToken')) is None
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document(self.__class__.auto_doc_id).get(field_paths=['released'], token=self.__class__.user.get('idToken')) == update_data

	def test_manual_doc_delete(self, ds):
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document('014').delete(self.__class__.user.get('idToken')) is None
		assert ds.collection('Marvels').document('Movies').collection('PhaseThree').document(self.__class__.auto_doc_id).delete(self.__class__.user.get('idToken')) is None

	def test_delete_test_user(self, auth):
		assert auth.delete_user_account(self.__class__.user.get('idToken'))


class TestFirestore:
	series1 = {
		'name': 'Loki',
		'lead': {'name': 'Tom Hiddleston'},
		'released': False,
		'year': 2021,
		'rating': 8.2,
		'prequel': None,
		'cast': ['Sophia Di Martino', 'Owen Wilson', 'Jonathan Majors', 'Wunmi Mosaku', 'Gugu Mbatha-Raw']
	}
	
	series2 = {
		'name': 'Moon Knight',
		'lead': {'name': 'Oscar Issac'},
		'released': False,
		'year': 2022,
		'rating': 7.4,
		'prequel': None,
		'cast': ['Ethan Hawke', 'May Calamawy', 'F. Murray Abraham']
	}

	auto_doc_id = None

	def test_manual_doc_set(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document('003').set(self.__class__.series1) is None

	def test_auto_doc_add(self, ds):
		doc_id = ds.collection('Marvels').document('Series').collection('PhaseFour').add(self.__class__.series2)
		assert doc_id

		self.__class__.auto_doc_id = doc_id

	def test_manual_doc_get(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document('003').get() == self.__class__.series1
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document(self.__class__.auto_doc_id).get() == self.__class__.series2

	def test_collection_get(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').get() == [{'003': self.__class__.series1}, {self.__class__.auto_doc_id: self.__class__.series2}]

	def test_manual_doc_get_filtered(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document('003').get(field_paths=['name']) == {'name': self.__class__.series1['name']}
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document(self.__class__.auto_doc_id).get(field_paths=['name']) == {'name': self.__class__.series2['name']}

	def test_collection_get_start_after(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('rating').start_after({'rating': 7.4}).get() == [{'003': self.__class__.series1}]
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('rating').start_after({'rating': 7.3}).get() == [{self.__class__.auto_doc_id: self.__class__.series2}, {'003': self.__class__.series1}]
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('rating').start_after({'rating': 8.5}).get() == []

	def test_collection_get_start_at(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('rating').start_at({'rating': 7.4}).get() == [{self.__class__.auto_doc_id: self.__class__.series2}, {'003': self.__class__.series1}]
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('rating').start_at({'rating': 8.0}).get() == [{'003': self.__class__.series1}]
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('rating').start_at({'rating': 8.5}).get() == []		

	def test_collection_get_select(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').select(['lead.name', 'released']).get() == [{'003': {'lead': self.__class__.series1['lead'], 'released': self.__class__.series1['released']}}, {self.__class__.auto_doc_id: {'lead': self.__class__.series2['lead'], 'released': self.__class__.series2['released']}}]

	def test_collection_get_offset(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('year').offset(1).get() == [{self.__class__.auto_doc_id: self.__class__.series2}]

	def test_collection_get_limit_to_first(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('year').limit_to_first(1).get() == [{'003': self.__class__.series1}]

	def test_collection_get_limit_to_last(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('year').limit_to_last(1).get() == [{'003': self.__class__.series1}]

	def test_collection_get_end_at(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('year').end_at({'year': 2010}).get() == []
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('year').end_at({'year': 2021}).get() == [{'003': self.__class__.series1}]

	def test_collection_get_end_before(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').order_by('year').end_before({'year': 2023}).get() == [{'003': self.__class__.series1}, {self.__class__.auto_doc_id: self.__class__.series2}]

	def test_collection_get_where(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').where('lead.name', 'in',  ['Benedict Cumberbatch', 'Robert Downey Jr.']).get() == []
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').where('rating', '<=', 8.0).get() == [{self.__class__.auto_doc_id: self.__class__.series2}]

	def test_manual_doc_update(self, ds):
		update_data = {'released': True}

		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document('003').update(update_data) is None
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document('003').get(field_paths=['released']) == update_data

		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document(self.__class__.auto_doc_id).update(update_data) is None
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document(self.__class__.auto_doc_id).get(field_paths=['released']) == update_data

	def test_manual_doc_delete(self, ds):
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document('003').delete() is None
		assert ds.collection('Marvels').document('Series').collection('PhaseFour').document(self.__class__.auto_doc_id).delete() is None




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


from google.oauth2.service_account import Credentials


def _service_account_creds_from_secret(service_account_secret):
	""" Service Account Credentials from Service Account Secrets.

	File path of the service account secret file
	in json format can  also be passed as the value
	of the parameter `service_account_secret`.

	:param service_account_secret: Service Account Secret Key from Firebase Console.
	:type service_account_secret: dict | str
	:return: Service Account Credentials
	:rtype: :class:`~google.oauth2.service_account.Credentials`
	"""

	credentials = None
	scopes = [
		'https://www.googleapis.com/auth/firebase.database',
		"https://www.googleapis.com/auth/datastore",
		'https://www.googleapis.com/auth/userinfo.email',
		"https://www.googleapis.com/auth/cloud-platform"
	]

	if type(service_account_secret) is str:
		credentials = Credentials.from_service_account_file(service_account_secret, scopes=scopes)
	if type(service_account_secret) is dict:
		credentials = Credentials.from_service_account_info(service_account_secret, scopes=scopes)

	return credentials




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


from collections import OrderedDict


class FirebaseKeyValue:

	def __init__(self, item):
		self.item = item

	def val(self):
		return self.item[1]

	def key(self):
		return self.item[0]


class FirebaseResponse:

	def __init__(self, firebases, query_key):
		self.firebases = firebases
		self.query_key = query_key

	def __getitem__(self, index):
		return self.firebases[index]

	def val(self):
		if isinstance(self.firebases, list):

			# unpack firebases into OrderedDict
			firebase_list = []

			# if firebase response was a list
			if isinstance(self.firebases[0].key(), int):

				for firebase in self.firebases:
					firebase_list.append(firebase.val())

				return firebase_list

			# if firebase response was a dict with keys
			for firebase in self.firebases:
				firebase_list.append((firebase.key(), firebase.val()))

			return OrderedDict(firebase_list)

		else:

			return self.firebases

	def key(self):
		return self.query_key

	def each(self):
		if isinstance(self.firebases, list):
			return self.firebases


def convert_to_firebase(items):
	firebase_list = []

	for item in items:
		firebase_list.append(FirebaseKeyValue(item))

	return firebase_list


def convert_list_to_firebase(items):
	firebase_list = []

	for item in items:
		firebase_list.append(FirebaseKeyValue([items.index(item), item]))

	return firebase_list




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


"""
A simple python wrapper for Google's `Firebase Cloud Storage REST API`_

.. _Firebase Cloud Storage REST API:
	https://firebase.google.com/docs/reference/rest/storage/rest
"""

import datetime
from google.cloud import storage
from urllib.parse import quote

from firebase._exception import raise_detailed_error


class Storage:
	""" Firebase Cloud Storage Service 

	:type credentials:
		:class:`~google.oauth2.service_account.Credentials`
	:param credentials: Service Account Credentials.

	:type requests: :class:`~requests.Session`
	:param requests: Session to make HTTP requests.

	:type storage_bucket: str
	:param storage_bucket: ``storageBucket`` from Firebase 
		configuration.
	"""

	def __init__(self, credentials, requests, storage_bucket):
		""" Constructor """

		self.credentials = credentials
		self.requests = requests
		self.storage_bucket = "https://firebasestorage.googleapis.com/v0/b/" + storage_bucket

		self.path = ""

		if credentials:
			client = storage.Client(credentials=credentials, project=storage_bucket)
			self.bucket = client.get_bucket(storage_bucket)

	def child(self, *args):
		""" Build paths to your storage.


		:type args: str
		:param args: Positional arguments to build path to storage.


		:return: A reference to the instance object.
		:rtype: Storage
		"""

		new_path = "/".join(args)

		if self.path:
			self.path += "/{}".format(new_path)
		else:
			if new_path.startswith("/"):
				new_path = new_path[1:]

			self.path = new_path

		return self

	def put(self, file, token=None):
		""" Upload file to storage.

		| For more details:
		| |upload_files|_

		.. |upload_files| replace::
			Firebase Documentation | Upload files with Cloud Storage on 
			Web

		.. _upload_files:
			https://firebase.google.com/docs/storage/web/upload-files#upload_files


		:type file: str
		:param file: Local path to file to upload.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.


		:return: Successful attempt returns :data:`None`.
		:rtype: :data:`None`
		"""

		# reset path
		path = self.path
		self.path = None

		if isinstance(file, str):
			file_object = open(file, 'rb')
		else:
			file_object = file

		request_ref = self.storage_bucket + "/o?name={0}".format(path)

		if token:
			headers = {"Authorization": "Firebase " + token}
			request_object = self.requests.post(request_ref, headers=headers, data=file_object)

			raise_detailed_error(request_object)

			return request_object.json()

		elif self.credentials:
			blob = self.bucket.blob(path)

			if isinstance(file, str):
				return blob.upload_from_filename(filename=file)
			else:
				return blob.upload_from_file(file_obj=file)

		else:
			request_object = self.requests.post(request_ref, data=file_object)

			raise_detailed_error(request_object)

			return request_object.json()

	def delete(self, token=None):
		""" Delete file from storage.

		| For more details:
		| |delete_a_file|_

		.. |delete_a_file| replace::
			Firebase Documentation | Delete files with Cloud Storage on 
			Web

		.. _delete_a_file:
			https://firebase.google.com/docs/storage/web/delete-files#delete_a_file


		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.
		"""

		# reset path
		path = self.path
		self.path = None

		# remove leading backlash
		if path.startswith('/'):
			path = path[1:]

		if self.credentials:
			self.bucket.delete_blob(path)
		else:
			request_ref = self.storage_bucket + "/o?name={0}".format(path)

			if token:
				headers = {"Authorization": "Firebase " + token}
				request_object = self.requests.delete(request_ref, headers=headers)
			else:
				request_object = self.requests.delete(request_ref)

			raise_detailed_error(request_object)

	def download(self, filename, token=None):
		""" Download file from storage.

		| For more details:
		| |download_data_via_url|_

		.. |download_data_via_url| replace::
			Firebase Documentation | Download files with Cloud Storage 
			on Web

		.. _download_data_via_url:
			https://firebase.google.com/docs/storage/web/download-files#download_data_via_url


		:type filename:  str
		:param filename: File name to be downloaded as.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults 
			to :data:`None`.
		"""

		if self.credentials:

			# reset path
			path = self.path
			self.path = None

			# remove leading backlash
			if path.startswith('/'):
				path = path[1:]

			blob = self.bucket.get_blob(path)
			if blob is not None:
				blob.download_to_filename(filename)

		elif token:
			headers = {"Authorization": "Firebase " + token}
			r = self.requests.get(self.get_url(token), stream=True, headers=headers)

			if r.status_code == 200:
				with open(filename, 'wb') as f:
					for chunk in r:
						f.write(chunk)

		else:
			r = self.requests.get(self.get_url(token), stream=True)

			if r.status_code == 200:
				with open(filename, 'wb') as f:
					for chunk in r:
						f.write(chunk)

	def get_url(self, token=None, expiration_hour=24):
		""" Fetches URL for file.


		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults
			to :data:`None`.

		:type expiration_hour: int
		:param expiration_hour: (Optional) time in ``hour`` for URL to
			expire after, defaults to 24 hours. Works only for links
			generated with admin credentials.

		:return: URL for the file.
		:rtype: str
		"""

		# reset path
		path = self.path
		self.path = None

		# remove leading backlash
		if path.startswith('/'):
			path = path[1:]

		if self.credentials:
			blob = self.bucket.get_blob(path)
			if blob:
				return blob.generate_signed_url(datetime.timedelta(hours=expiration_hour), method='GET')

		elif token:

			# retrieve download tokens first
			headers = {"Authorization": "Firebase " + token}
			request_ref = "{0}/o/{1}".format(self.storage_bucket, quote(path, safe=''))
			request_object = self.requests.get(request_ref, headers=headers)

			raise_detailed_error(request_object)

			return "{0}/o/{1}?alt=media&token={2}".format(self.storage_bucket, quote(path, safe=''), request_object.json()['downloadTokens'])

		return "{0}/o/{1}?alt=media".format(self.storage_bucket, quote(path, safe=''))

	def list_files(self):
		""" List of all files in storage.

		| for more details:
		| |list_all_files|_

		.. |list_all_files| replace::
			Firebase Documentation | List files with Cloud Storage on 
			Web

		.. _list_all_files:
			https://firebase.google.com/docs/storage/web/list-files#list_all_files


		:return: list of :class:`~gcloud.storage.blob.Blob`
		:rtype: :class:`~gcloud.storage.bucket._BlobIterator`
		"""

		return self.bucket.list_blobs()




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


"""
A simple python wrapper for Google's `Firebase Cloud Firestore REST API`_

.. _Firebase Cloud Firestore REST API:
	https://firebase.google.com/docs/firestore/reference/rest
"""

from math import ceil
from proto.message import Message
from google.cloud.firestore import Client
from google.cloud.firestore_v1._helpers import *
from google.cloud.firestore_v1.query import Query
from google.cloud.firestore_v1.collection import CollectionReference
from google.cloud.firestore_v1.base_query import _enum_from_direction

from ._utils import _from_datastore, _to_datastore
from firebase._exception import raise_detailed_error


class Firestore:
	""" Firebase Firestore Service

	:type api_key: str
	:param api_key: ``apiKey`` from Firebase configuration

	:type credentials: :class:`~google.oauth2.service_account.Credentials`
	:param credentials: Service Account Credentials

	:type project_id: str
	:param project_id: ``projectId`` from Firebase configuration

	:type requests: :class:`~requests.Session`
	:param requests: Session to make HTTP requests
	"""

	def __init__(self, api_key, credentials, project_id, requests):
		""" Constructor method """

		self._api_key = api_key
		self._credentials = credentials
		self._project_id = project_id
		self._requests = requests

	def collection(self, collection_id):
		""" Get reference to a collection in a Firestore database.


		:type collection_id: str
		:param collection_id: An ID of collection in firestore.


		:return: Reference to a collection.
		:rtype: Collection
		"""

		return Collection([collection_id], api_key=self._api_key, credentials=self._credentials, project_id=self._project_id, requests=self._requests)


class Collection:
	""" A reference to a collection in a Firestore database.

	:type collection_path: list
	:param collection_path: Collective form of strings to create a
		Collection.

	:type api_key: str
	:param api_key: ``apiKey`` from Firebase configuration

	:type credentials: :class:`~google.oauth2.service_account.Credentials`
	:param credentials: Service Account Credentials

	:type project_id: str
	:param project_id: ``projectId`` from Firebase configuration

	:type requests: :class:`~requests.Session`
	:param requests: Session to make HTTP requests
	"""

	def __init__(self, collection_path, api_key, credentials, project_id, requests):
		""" Constructor method """

		self._path = collection_path

		self._api_key = api_key
		self._credentials = credentials
		self._project_id = project_id
		self._requests = requests

		self._base_path = f"projects/{self._project_id}/databases/(default)/documents"
		self._base_url = f"https://firestore.googleapis.com/v1/{self._base_path}"

		if self._credentials:
			self.__datastore = Client(credentials=self._credentials, project=self._project_id)

		self._query = {}
		self._is_limited_to_last = False

	def _build_query(self):
		""" Builds query for firestore to execute.


		:return: An query.
		:rtype: :class:`~google.cloud.firestore_v1.query.Query`
		"""

		if self._credentials:
			_query = _build_db(self.__datastore, self._path)
		else:
			_query = Query(CollectionReference(self._path.pop()))

		for key, val in self._query.items():
			if key == 'endAt':
				_query = _query.end_at(val)
			elif key == 'endBefore':
				_query = _query.end_before(val)
			elif key == 'limit':
				_query = _query.limit(val)
			elif key == 'limitToLast':
				_query = _query.limit_to_last(val)
			elif key == 'offset':
				_query = _query.offset(val)
			elif key == 'orderBy':
				for q in val:
					_query = _query.order_by(q[0], **q[1])
			elif key == 'select':
				_query = _query.select(val)
			elif key == 'startAfter':
				_query = _query.start_after(val)
			elif key == 'startAt':
				_query = _query.start_at(val)
			elif key == 'where':
				for q in val:
					_query = _query.where(q[0], q[1], q[2])

		if not self._credentials and _query._limit_to_last:

			self._is_limited_to_last = _query._limit_to_last

			for order in _query._orders:
				order.direction = _enum_from_direction(
					_query.DESCENDING
					if order.direction == _query.ASCENDING
					else _query.ASCENDING
				)

			_query._limit_to_last = False

		self._path.clear()
		self._query.clear()

		return _query

	def add(self, data, token=None):
		""" Create a document in the Firestore database with the
		provided data using an auto generated ID for the document.


		:type data: dict
		:param data: Data to be stored in firestore.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults
			to :data:`None`.


		:return: returns the auto generated document ID, used to store
			the data.
		:rtype: str
		"""

		path = self._path.copy()
		self._path.clear()

		if self._credentials:
			db_ref = _build_db(self.__datastore, path)

			response = db_ref.add(data)

			return response[1].id

		else:
			req_ref = f"{self._base_url}/{'/'.join(path)}?key={self._api_key}"

			if token:
				headers = {"Authorization": "Firebase " + token}
				response = self._requests.post(req_ref, headers=headers, json=_to_datastore(data))

			else:
				response = self._requests.post(req_ref, json=_to_datastore(data))

			raise_detailed_error(response)

			doc_id = response.json()['name'].split('/')

			return doc_id.pop()

	def document(self, document_id):
		""" A reference to a document in a collection.


		:type document_id: str
		:param document_id: An ID of document inside a collection.


		:return: Reference to a document.
		:rtype: Document
		"""

		self._path.append(document_id)
		return Document(self._path, api_key=self._api_key, credentials=self._credentials, project_id=self._project_id, requests=self._requests)

	def end_at(self, document_fields):
		""" End query at a cursor with this collection as parent.


		:type document_fields: dict
		:param document_fields: A dictionary of fields representing a
			query results cursor. A cursor is a collection of values
			that represent a position in a query result set.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		self._query['endAt'] = document_fields

		return self

	def end_before(self, document_fields):
		""" End query before a cursor with this collection as parent.


		:type document_fields: dict
		:param document_fields: A dictionary of fields representing a
			query results cursor. A cursor is a collection of values
			that represent a position in a query result set.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		self._query['endBefore'] = document_fields

		return self

	def get(self, token=None):
		""" Returns a list of dict's containing document ID and the
		data stored within them.


		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults
			to :data:`None`.


		:return: A list of document ID's with the data they possess.
		:rtype: list
		"""

		docs = []

		if self._credentials:
			db_ref = self._build_query()

			results = db_ref.get()

			for result in results:
				docs.append({result.id: result.to_dict()})

		else:

			body = None

			if len(self._query) > 0:
				req_ref = f"{self._base_url}/{'/'.join(self._path[:-1])}:runQuery?key={self._api_key}"

				body = {
					"structuredQuery": json.loads(Message.to_json(self._build_query()._to_protobuf()))
				}

			else:
				req_ref = f"{self._base_url}/{'/'.join(self._path)}?key={self._api_key}"

			if token:
				headers = {"Authorization": "Firebase " + token}

				if body:
					response = self._requests.post(req_ref, headers=headers, json=body)
				else:
					response = self._requests.get(req_ref, headers=headers)

			else:

				if body:
					response = self._requests.post(req_ref, json=body)
				else:
					response = self._requests.get(req_ref)

			raise_detailed_error(response)

			if isinstance(response.json(), dict):
				for doc in response.json()['documents']:
					doc_id = doc['name'].split('/')
					docs.append({doc_id.pop(): _from_datastore({'fields': doc['fields']})})

			elif isinstance(response.json(), list):
				for doc in response.json():
					fields = {}

					if doc.get('document'):

						if doc.get('document').get('fields'):
							fields = doc['document']['fields']

						doc_id = doc['document']['name'].split('/')
						docs.append({doc_id.pop(): _from_datastore({'fields': fields})})

			if self._is_limited_to_last:
				docs = list(reversed(list(docs)))

		return docs

	def limit_to_first(self, count):
		""" Create a limited query with this collection as parent.

			.. note::
				`limit_to_first` and `limit_to_last` are mutually
				exclusive. Setting `limit_to_first` will drop
				previously set `limit_to_last`.


		:type count: int
		:param count: Maximum number of documents to return that match
			the query.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		self._query['limit'] = count

		return self

	def limit_to_last(self, count):
		""" Create a limited to last query with this collection as
		parent.

			.. note::
				`limit_to_first` and `limit_to_last` are mutually
				exclusive. Setting `limit_to_first` will drop
				previously set `limit_to_last`.


		:type count: int
		:param count: Maximum number of documents to return that
			match the query.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		self._query['limitToLast'] = count

		return self

	def offset(self, num_to_skip):
		""" Skip to an offset in a query with this collection as parent.


		:type num_to_skip: int
		:param num_to_skip: The number of results to skip at the
			beginning of query results. (Must be non-negative.)


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		self._query['offset'] = num_to_skip

		return self

	def order_by(self, field_path, **kwargs):
		""" Create an "order by" query with this collection as parent.


		:type field_path: str
		:param field_path: A field path (``.``-delimited list of field
			names) on which to order the query results.

		:Keyword Arguments:
			* *direction* ( :class:`str` ) --
				Sort query results in ascending/descending order on a field.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		arr = []

		if self._query.get('orderBy'):
			arr = self._query['orderBy']

		arr.append([field_path, kwargs])

		self._query['orderBy'] = arr

		return self

	def select(self, field_paths):
		""" Create a "select" query with this collection as parent.

		:type field_paths: list
		:param field_paths: A list of field paths (``.``-delimited list
			of field names) to use as a projection of document fields
			in the query results.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		self._query['select'] = field_paths

		return self

	def start_after(self, document_fields):
		""" Start query after a cursor with this collection as parent.


		:type document_fields: dict
		:param document_fields: A dictionary of fields representing
			a query results cursor. A cursor is a collection of values
			that represent a position in a query result set.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		self._query['startAfter'] = document_fields

		return self

	def start_at(self, document_fields):
		""" Start query at a cursor with this collection as parent.


		:type document_fields: dict
		:param document_fields: A dictionary of fields representing a
			query results cursor. A cursor is a collection of values
			that represent a position in a query result set.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		self._query['startAt'] = document_fields

		return self

	def where(self, field_path, op_string, value):
		""" Create a "where" query with this collection as parent.


		:type field_path: str
		:param field_path: A field path (``.``-delimited list of field
			names) for the field to filter on.

		:type op_string: str
		:param op_string: A comparison operation in the form of a
			string. Acceptable values are ``<``, ``<=``, ``==``, ``!=``
			, ``>=``, ``>``, ``in``, ``not-in``, ``array_contains`` and
			``array_contains_any``.

		:type value: Any
		:param value: The value to compare the field against in the
			filter. If ``value`` is :data:`None` or a NaN, then ``==``
			is the only allowed operation.  If ``op_string`` is ``in``,
			``value`` must be a sequence of values.


		:return: A reference to the instance object.
		:rtype: Collection
		"""

		arr = []

		if self._query.get('where'):
			arr = self._query['where']

		arr.append([field_path, op_string, value])

		self._query['where'] = arr

		return self


class Document:
	""" A reference to a document in a Firestore database.

	:type document_path: list
	:param document_path: Collective form of strings to create a
		Document.

	:type api_key: str
	:param api_key: ``apiKey`` from Firebase configuration

	:type credentials: :class:`~google.oauth2.service_account.Credentials`
	:param credentials: Service Account Credentials

	:type project_id: str
	:param project_id: ``projectId`` from Firebase configuration

	:type requests: :class:`~requests.Session`
	:param requests: Session to make HTTP requests
	"""

	def __init__(self, document_path, api_key, credentials, project_id, requests):
		""" Constructor method """

		self._path = document_path

		self._api_key = api_key
		self._credentials = credentials
		self._project_id = project_id
		self._requests = requests

		self._base_path = f"projects/{self._project_id}/databases/(default)/documents"
		self._base_url = f"https://firestore.googleapis.com/v1/{self._base_path}"

		if self._credentials:
			self.__datastore = Client(credentials=self._credentials, project=self._project_id)

	def collection(self, collection_id):
		""" A reference to a collection in a Firestore database.


		:type collection_id: str
		:param collection_id: An ID of collection in firestore.


		:return: Reference to a collection.
		:rtype: Collection
		"""

		self._path.append(collection_id)
		return Collection(self._path, api_key=self._api_key, credentials=self._credentials, project_id=self._project_id, requests=self._requests)

	def delete(self, token=None):
		""" Deletes the current document from firestore.

		| For more details:
		| |delete_documents|_

		.. |delete_documents| replace::
			Firebase Documentation | Delete data from Cloud
			Firestore | Delete documents

		.. _delete_documents:
			https://firebase.google.com/docs/firestore/manage-data/delete-data#delete_documents

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults
			to :data:`None`.
		"""

		path = self._path.copy()
		self._path.clear()

		if self._credentials:
			db_ref = _build_db(self.__datastore, path)

			db_ref.delete()

		else:
			req_ref = f"{self._base_url}/{'/'.join(path)}?key={self._api_key}"

			if token:
				headers = {"Authorization": "Firebase " + token}
				response = self._requests.delete(req_ref, headers=headers)

			else:
				response = self._requests.delete(req_ref)

			raise_detailed_error(response)

	def get(self, field_paths=None, token=None):
		""" Read data from a document in firestore.


		:type field_paths: list
		:param field_paths: (Optional) A list of field paths
			(``.``-delimited list of field names) to filter data, and
			return the filtered values only, defaults
			to :data:`None`.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults
			to :data:`None`.


		:return: The whole data stored in the document unless filtered
			to retrieve specific fields.
		:rtype: dict
		"""

		path = self._path.copy()
		self._path.clear()

		if self._credentials:
			db_ref = _build_db(self.__datastore, path)

			result = db_ref.get(field_paths=field_paths)

			return result.to_dict()

		else:

			mask = ''

			if field_paths:
				for field_path in field_paths:
					mask = f"{mask}mask.fieldPaths={field_path}&"

			req_ref = f"{self._base_url}/{'/'.join(path)}?{mask}key={self._api_key}"

			if token:
				headers = {"Authorization": "Firebase " + token}
				response = self._requests.get(req_ref, headers=headers)

			else:
				response = self._requests.get(req_ref)

			raise_detailed_error(response)

			return _from_datastore(response.json())

	def set(self, data, token=None):
		""" Add data to a document in firestore.

		| For more details:
		| |set_a_document|_

		.. |set_a_document| replace::
			Firebase Documentation | Add data to Cloud Firestore | Set
			a document

		.. _set_a_document:
			https://firebase.google.com/docs/firestore/manage-data/add-data#set_a_document


		:type data: dict
		:param data: Data to be stored in firestore.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults
			to :data:`None`.
		"""

		path = self._path.copy()
		self._path.clear()

		if self._credentials:
			db_ref = _build_db(self.__datastore, path)

			db_ref.set(data)

		else:

			req_ref = f"{self._base_url}:commit?key={self._api_key}"

			body = {
				"writes": [
					Message.to_dict(pbs_for_set_no_merge(f"{self._base_path}/{'/'.join(path)}", data)[0])
					]
			}

			if token:
				headers = {"Authorization": "Firebase " + token}
				response = self._requests.post(req_ref, headers=headers, json=body)

			else:
				response = self._requests.post(req_ref, json=body)

			raise_detailed_error(response)

	def update(self, data, token=None):
		""" Update stored data inside a document in firestore.


		:type data: dict
		:param data: Data to be stored in firestore.

		:type token: str
		:param token: (Optional) Firebase Auth User ID Token, defaults
			to :data:`None`.
		"""

		path = self._path.copy()
		self._path.clear()

		if self._credentials:
			db_ref = _build_db(self.__datastore, path)

			db_ref.update(data)

		else:
			req_ref = f"{self._base_url}:commit?key={self._api_key}"

			body = {
				"writes": [
					Message.to_dict(pbs_for_update(f"{self._base_path}/{'/'.join(path)}", data, None)[0])
				]
			}

			if token:
				headers = {"Authorization": "Firebase " + token}
				response = self._requests.post(req_ref, headers=headers, json=body)

			else:
				response = self._requests.post(req_ref, json=body)

			raise_detailed_error(response)


def _build_db(db, path):
	""" Returns a reference to Collection/Document with admin
	credentials.


	:type db: :class:`~google.cloud.firestore.Client`
	:param db: Reference to Firestore Client.

	:type path: list
	:param path: Collective form of strings to create a document.


	:return: Reference to collection/document to perform CRUD 
		operations.
	:rtype: :class:`~google.cloud.firestore_v1.document.CollectionReference`
		or :class:`~google.cloud.firestore_v1.document.DocumentReference`
	"""

	n = ceil(len(path) / 2)

	for _ in range(n):
		db = db.collection(path.pop(0))

		if len(path) > 0:
			db = db.document(path.pop(0))

	return db




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


"""
A simple python wrapper for Google's
`Firebase Authentication REST API`_

.. _Firebase Authentication REST API: https://firebase.google.com/docs/reference/rest/auth
"""

import json
import math
import datetime
import python_jwt as jwt
import jwcrypto.jwk as jwk
from urllib.parse import parse_qs
from cryptography.hazmat.primitives.serialization import Encoding, NoEncryption, PrivateFormat

from firebase._exception import raise_detailed_error


class Auth:
	""" Firebase Authentication Service

	:type api_key: str
	:param api_key: ``apiKey`` from Firebase configuration

	:type credentials: :class:`~google.oauth2.service_account.Credentials`
	:param credentials: Service Account Credentials

	:type requests: :class:`~requests.Session`
	:param requests: Session to make HTTP requests

	:type client_secret: str or dict
	:param client_secret: (Optional) File path to or the dict object
		from social client secret file, defaults to :data:`None`.

	"""

	def __init__(self, api_key, credentials, requests, client_secret=None):
		""" Constructor method """

		self.api_key = api_key
		self.credentials = credentials
		self.requests = requests

		self.provider_id = None
		self.session_id = None

		if client_secret:
			self.client_secret = _load_client_secret(client_secret)

	def authenticate_login_with_google(self):
		""" Redirect the user to Google's OAuth 2.0 server to initiate 
		the authentication and authorization process.

		:return: Google Sign In URL
		:rtype: str
		"""
		return self.create_authentication_uri('google.com')

	def create_authentication_uri(self, provider_id):
		""" Creates an authentication URI for the given social
		provider.

		| For more details:
		| |section-fetch-providers-for-email|_

		.. |section-fetch-providers-for-email| replace::
			Firebase Auth REST API | Fetch providers for email

		.. _section-fetch-providers-for-email:
			https://firebase.google.com/docs/reference/rest/auth#section-fetch-providers-for-email


		:type provider_id: str
		:param provider_id: The IdP ID. For white listed IdPs it's a
			short domain name e.g. 'google.com', 'aol.com', 'live.net'
			and 'yahoo.com'. For other OpenID IdPs it's the OP
			identifier.


		:return: The URI used by the IDP to authenticate the user.
		:rtype: str
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/createAuthUri?key={0}".format(self.api_key)

		data = {
			"authFlowType": 'CODE_FLOW',
			"clientId": self.client_secret['client_id'],
			"providerId": provider_id,
			"continueUri": self.client_secret['redirect_uris'][0],
			"customParameter": {
				"access_type": 'offline',
				"prompt": 'select_account',
				"include_granted_scopes": 'true',
			}
		}

		headers = {"content-type": "application/json; charset=UTF-8"}
		request_object = self.requests.post(request_ref, headers=headers, json=data)

		raise_detailed_error(request_object)

		self.provider_id = provider_id
		self.session_id = request_object.json()['sessionId']

		return request_object.json()['authUri']

	def sign_in_with_email_and_password(self, email, password):
		""" Sign in a user with an email and password.

		| For more details:
		| `Firebase Auth REST API | section-sign-in-email-password`_

		.. _Firebase Auth REST API | section-sign-in-email-password: https://firebase.google.com/docs/reference/rest/auth#section-sign-in-email-password

		:type email: str
		:param email: The email the user is signing in with.

		:type password: str
		:param password: The password for the account.

		:return: UserInfo and Firebase Auth Tokens.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/verifyPassword?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"email": email, "password": password, "returnSecureToken": True})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return _token_expire_time(request_object.json())

	def sign_in_anonymous(self):
		""" Sign In Anonymously.

		| For more details:
		| `Firebase Auth REST API | section-sign-in-anonymously`_

		.. _Firebase Auth REST API | section-sign-in-anonymously: https://firebase.google.com/docs/reference/rest/auth#section-sign-in-anonymously

		:return: Firebase Auth Tokens.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/signupNewUser?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"returnSecureToken": True})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return _token_expire_time(request_object.json())

	def create_custom_token(self, uid, additional_claims=None, expiry_minutes=60):
		""" Create a Firebase Auth custom token.

		| For more details:
		| `Firebase Documentation | Create Custom tokens`_

		.. _Firebase Documentation | Create Custom tokens: https://firebase.google.com/docs/auth/admin/create-custom-tokens

		:type uid: str
		:param uid: The unique identifier of the user, must be a
			string, between 1-36 characters long.

		:type additional_claims:  dict or None
		:param additional_claims: Optional custom claims to include
			in the Security Rules ``auth`` / ``request.auth``
			variables.

		:type expiry_minutes:  int
		:param expiry_minutes: The time, in minutes since the UNIX
			epoch, at which the token expires.
			Default value is 60.

		:return: Firebase Auth custom token.
		:rtype: str
		"""

		service_account_email = self.credentials.service_account_email
		private_key = jwk.JWK.from_pem(self.credentials.signer._key.private_bytes(encoding=Encoding.PEM, format=PrivateFormat.PKCS8, encryption_algorithm=NoEncryption()))

		payload = {
			"iss": service_account_email,
			"sub": service_account_email,
			"aud": "https://identitytoolkit.googleapis.com/google.identity.identitytoolkit.v1.IdentityToolkit",
			"uid": uid
		}

		if additional_claims:
			payload["claims"] = additional_claims

		exp = datetime.timedelta(minutes=expiry_minutes)

		return jwt.generate_jwt(payload, private_key, "RS256", exp, other_headers={'kid': self.credentials.signer._key_id})

	def sign_in_with_custom_token(self, token):
		""" Exchange custom token for an ID and refresh token.

		| For more details:
		| `Firebase Auth REST API | section-verify-custom-token`_

		.. _Firebase Auth REST API | section-verify-custom-token : https://firebase.google.com/docs/reference/rest/auth#section-verify-custom-token

		:type token: str
		:param token: A Firebase Auth custom token from which to
			create an ID and refresh token pair.

		:return: Firebase Auth Tokens.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/verifyCustomToken?key={0}".format(self.api_key)	# noqa

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"returnSecureToken": True, "token": token})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return _token_expire_time(request_object.json())

	def refresh(self, refresh_token):
		""" Refresh a Firebase ID token.

		| For more details:
		| `Firebase Auth REST API | section-refresh-token`_

		.. _Firebase Auth REST API | section-refresh-token : https://firebase.google.com/docs/reference/rest/auth#section-refresh-token

		:type refresh_token: str
		:param refresh_token: A Firebase Auth refresh token.

		:return: New (Refreshed) Firebase Auth tokens for the account.
		:rtype: dict
		"""

		request_ref = "https://securetoken.googleapis.com/v1/token?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"grantType": "refresh_token", "refreshToken": refresh_token})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)
		request_object_json = request_object.json()

		# handle weirdly formatted response
		user = {
			"localId": request_object_json["user_id"],
			"idToken": request_object_json["id_token"],
			"refreshToken": request_object_json["refresh_token"],
			"expiresIn": request_object_json["expires_in"]
		}

		return _token_expire_time(user)

	def get_account_info(self, id_token):
		""" Fetch user's stored account information.

		| For more details:
		| `Firebase Auth REST API | section-get-account-info`_

		.. _Firebase Auth REST API | section-get-account-info : https://firebase.google.com/docs/reference/rest/auth#section-get-account-info

		:type id_token: str
		:param id_token: The Firebase ID token of the account.

		:return: The account info, associated with the given
			Firebase ID token.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/getAccountInfo?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"idToken": id_token})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return request_object.json()

	def send_email_verification(self, id_token):
		""" Send an email verification to verify email ownership.

		| For more details:
		| `Firebase Auth REST API | section-send-email-verification`_

		.. _Firebase Auth REST API | section-send-email-verification : https://firebase.google.com/docs/reference/rest/auth#section-send-email-verification

		:type id_token: str
		:param id_token: The Firebase ID token of the user to verify.

		:return: The email of the account associated with Firebase ID
			token.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/getOobConfirmationCode?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"requestType": "VERIFY_EMAIL", "idToken": id_token})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return request_object.json()

	def send_password_reset_email(self, email):
		""" Send a password reset email.

		| For more details:
		| `Firebase Auth REST API | section-send-password-reset-email`_

		.. _Firebase Auth REST API | section-send-password-reset-email: https://firebase.google.com/docs/reference/rest/auth#section-send-password-reset-email

		:type email: str
		:param email: User's email address.

		:return: User's email address.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/getOobConfirmationCode?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"requestType": "PASSWORD_RESET", "email": email})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return request_object.json()

	def verify_password_reset_code(self, reset_code, new_password):
		""" Reset password using code.

		| For more details:
		| `Firebase Auth REST API | #section-confirm-reset-password`_

		.. _Firebase Auth REST API | #section-confirm-reset-password: https://firebase.google.com/docs/reference/rest/auth#section-confirm-reset-password

		:type reset_code: str
		:param reset_code: The email action code sent to the user's
			email for resetting the password.

		:type new_password: str
		:param new_password: The user's new password.

		:return: User Email and Type of the email action code.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/resetPassword?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"oobCode": reset_code, "newPassword": new_password})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return request_object.json()

	def create_user_with_email_and_password(self, email, password):
		""" Create a new user with email and password.

		| For more details:
		| `Firebase Auth REST API | section-create-email-password`_

		.. _Firebase Auth REST API | section-create-email-password: https://firebase.google.com/docs/reference/rest/auth#section-create-email-password

		:type email:  str
		:param email: The email for the user to create.

		:type password:  str
		:param password: The password for the user to create.

		:return: User Email and Firebase Auth Tokens.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/signupNewUser?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"email": email, "password": password, "returnSecureToken": True})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return request_object.json()

	def delete_user_account(self, id_token):
		""" Delete an existing user.

		| For more details:
		| `Firebase Auth REST API | section-delete-account`_

		.. _Firebase Auth REST API | section-delete-account: https://firebase.google.com/docs/reference/rest/auth#section-delete-account

		:type id_token: str
		:param id_token: The Firebase ID token of the user to
			delete.
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/deleteAccount?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"idToken": id_token})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return request_object.json()

	def sign_in_with_oauth_credential(self, oauth2callback_url):
		""" Sign In With OAuth credential.

		| For more details:
		| |section-sign-in-with-oauth-credential|_

		.. |section-sign-in-with-oauth-credential| replace::
			Firebase Auth REST API | Sign in with OAuth credential

		.. _section-sign-in-with-oauth-credential:
			https://firebase.google.com/docs/reference/rest/auth#section-sign-in-with-oauth-credential


		:type oauth2callback_url: str
		:param oauth2callback_url: The URL redirected to after
			successful authorization from the provider.

		:return: User account info and Firebase Auth Tokens.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/verifyAssertion?key={0}".format(self.api_key)

		token = self._token_from_auth_url(oauth2callback_url)
		data = {
			'postBody': 'providerId={0}&{1}={2}'.format(self.provider_id, token['type'], token['value']),
			'autoCreate': 'true',
			'requestUri': self.client_secret['redirect_uris'][0],
			'sessionId': self.session_id,
			'returnSecureToken': 'true',
			'returnRefreshToken': 'true',
			'returnIdpCredential': 'false',
		}

		headers = {"content-type": "application/json; charset=UTF-8"}
		request_object = self.requests.post(request_ref, headers=headers, json=data)

		raise_detailed_error(request_object)

		return _token_expire_time(request_object.json())

	def _token_from_auth_url(self, url):
		""" Fetch tokens using the authorization code from given URL.


		:type url: str
		:param url: The URL redirected to after successful
			authorization from the provider.


		:return: The OAuth credential (an ID token).
		:rtype: dict
		"""

		request_ref = 'https://www.googleapis.com/oauth2/v4/token'

		auth_url_values = parse_qs(url[url.index('?') + 1:])

		data = {
			'client_id': self.client_secret['client_id'],
			'client_secret': self.client_secret['client_secret'],
			'code': auth_url_values['code'][0],
			'grant_type': 'authorization_code',
			'redirect_uri': self.client_secret['redirect_uris'][0],
		}

		headers = {"content-type": "application/x-www-form-urlencoded; charset=UTF-8"}
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return {
			'type': 'id_token',
			'value': request_object.json()['id_token'],
		}

	def update_profile(self, id_token, display_name=None, photo_url=None, delete_attribute=None):
		""" Update a user's profile (display name / photo URL).

		| For more details:
		| `Firebase Auth REST API | section-update-profile`_

		.. _Firebase Auth REST API | section-update-profile: https://firebase.google.com/docs/reference/rest/auth#section-update-profile

		:type id_token: str
		:param id_token: A Firebase Auth ID token for the user.

		:type display_name: str or None
		:param display_name: User's new display name.

		:type photo_url: None or str
		:param photo_url: User's new photo url.

		:type delete_attribute: list[str] or None
		:param delete_attribute: List of attributes to delete,
			"DISPLAY_NAME" or "PHOTO_URL". This will nullify these
			values.

		:return: UserInfo and Firebase Auth Tokens.
		:rtype: dict
		"""

		request_ref = "https://www.googleapis.com/identitytoolkit/v3/relyingparty/setAccountInfo?key={0}".format(self.api_key)

		headers = {"content-type": "application/json; charset=UTF-8"}
		data = json.dumps({"idToken": id_token, "displayName": display_name, "photoURL": photo_url, "deleteAttribute": delete_attribute, "returnSecureToken": True})
		request_object = self.requests.post(request_ref, headers=headers, data=data)

		raise_detailed_error(request_object)

		return request_object.json()


def _load_client_secret(secret):
	""" Load social providers' client secret from file if file path
	provided.

	This function also restructures the dict object to make it
	compatible for usage.


	:type secret: str or dict
	:param secret: File path to or the dict object from social client
		secret file.

	:return: social client secret
	:rtype: dict
	"""

	if type(secret) is str:
		with open(secret) as file:
			secret = json.load(file)

	# Google client secrets are stored within 'web' key
	# We will remove the key, and replace it with the dict type value of it
	secret = secret['web']

	return secret


def _token_expire_time(user):
	""" Adds expire time of the token in the token dictionary.

	For safety purposes, token is set to expire after 59mins instead
	of an hour expiry time when received.

	:type user: dict
	:param user: The token dictionary received after signing in users.
	
	:return: Token dictionary with an ``expiresAt`` key.
	:rtype: dict
	"""
	
	user['expiresAt'] = math.floor(datetime.datetime.today().timestamp() + int(user.get('expiresIn')) - 60)
	
	return user




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


# get this from firebase console
# go to project settings, general tab and click "Add Firebase to your web app"
SIMPLE_CONFIG = {
	"apiKey": "",
	"authDomain": "",
	"databaseURL": "",
	"storageBucket": "",
}

# get this json file from firebase console
# go to project settings, service accounts tab and click generate new private key
SERVICE_ACCOUNT_PATH = "firebase-adminsdk.json"

SERVICE_CONFIG = dict(SIMPLE_CONFIG, serviceAccount=SERVICE_ACCOUNT_PATH)




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


from requests import Session
from requests import adapters


def _custom_request(max_retries=None):
	""" Custom Session with N retries.

	Incase a request was not completed successfully due to minor
	errors, such as connection failures due to unreliable networks;
	or server side overload to proceed the request, retrying the
	request couple of times might solve it, without raising an error.

	if no value is sent through the function,
	`max_retries` is set to 3.

	:param max_retries: number of retries.
	:type max_retries: int | None
	:return: custom session
	:rtype: Session
	"""

	session = Session()
	adapter = adapters.HTTPAdapter(max_retries=max_retries)

	for scheme in ('http://', 'https://'):
		session.mount(scheme, adapter)

	return session




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


from datetime import datetime
from base64 import b64encode, b64decode
from google.protobuf.json_format import MessageToDict
from google.cloud.firestore_v1._helpers import GeoPoint
from google.api_core.datetime_helpers import DatetimeWithNanoseconds


def _from_datastore(data):
	""" Converts a map of Firestore ``data``-s to Python dictionary.


	:type data: dict
	:param data: A map of firestore data.


	:return: A dictionary of native Python values converted
		from the ``data``.
	:rtype: dict
	"""

	data_to_restructure = data['fields']

	for key, val in data_to_restructure.items():

		if val.get('mapValue'):
			data_to_restructure[key] = _from_datastore(val['mapValue'])

		elif val.get('arrayValue'):
			arr = []

			for x in val['arrayValue']['values']:
				arr.append(_decode_datastore(x))

			data_to_restructure[key] = arr

		else:
			data_to_restructure[key] = _decode_datastore(val)

	return data_to_restructure


def _decode_datastore(value):
	""" Converts a Firestore ``value`` to a native Python value.


	:type value: dict
	:param value: A Firestore data to be decoded / parsed /
		converted.


	:return: A native Python value converted from the ``value``.
	:rtype: :data:`None` or :class:`bool` or :class:`bytes`
		or :class:`int` or :class:`float` or :class:`str` or
		:class:`dict` or
		:class:`~google.api_core.datetime_helpers.DatetimeWithNanoseconds`
		or  :class:`~google.cloud.firestore_v1._helpers.GeoPoint`.

	:raises TypeError: For value types that are unsupported.
	"""

	if value.get('nullValue', False) is None:
		return value['nullValue']

	elif value.get('booleanValue') is not None:
		return bool(value['booleanValue'])

	elif value.get('bytesValue'):
		return b64decode(value['bytesValue'].encode('utf-8'))

	elif value.get('integerValue'):
		return int(value['integerValue'])

	elif value.get('doubleValue'):
		return float(value['doubleValue'])

	elif value.get('stringValue'):
		return str(value['stringValue'])

	elif value.get('mapValue'):
		return _from_datastore(value['mapValue'])

	elif value.get('timestampValue'):
		return DatetimeWithNanoseconds.from_rfc3339(value['timestampValue'])

	elif value.get('geoPointValue'):
		return GeoPoint(float(value['timestampValue']['latitude']), float(value['timestampValue']['longitude']))

	else:
		raise TypeError("Cannot convert to a Python Value", value, "Invalid type", type(value))


def _to_datastore(data):
	""" Converts a Python dictionary ``data``-s to map of Firestore.


	:type data: dict
	:param data: A Python dictionary containing data.


	:return: A map of Firebase values converted from the ``data``.
	:rtype: dict

	:raises ValueError: Raised when a key in the python dictionary is
		Non-alphanum char without *`* (ticks) at start and end.
	"""

	restructured_data = {}

	for key, val in data.items():

		if ' ' in key and (not key.startswith('`') or not key.endswith('`')):
			raise ValueError(f'Non-alphanum char in element with leading alpha: {key}')

		key = str(key)

		if isinstance(val, dict):
			restructured_data[key] = {'mapValue': _to_datastore(val)}

		elif isinstance(val, list):
			arr = []

			for x in val:
				arr.append(_encode_datastore_value(x))

			restructured_data[key] = {'arrayValue': {'values': arr}}

		else:
			restructured_data[key] = _encode_datastore_value(val)

	return {'fields': restructured_data}


def _encode_datastore_value(value):
	""" Converts a Python ``value`` to a Firebase value.


	:type value: :data:`None` or :class:`bool` or :class:`bytes`
		or :class:`int` or :class:`float` or :class:`str` or
		:class:`dict` or :class:`~datetime.datetime` or
		:class:`~google.api_core.datetime_helpers.DatetimeWithNanoseconds`
		or  :class:`~google.cloud.firestore_v1._helpers.GeoPoint`.
	:param value: A Python data to be encoded/converted to Firebase.


	:return: A Firebase value converted from ``value``.
	:rtype: dict

	:raises TypeError: Raised when unsupported data type given.
	"""

	if value is None:
		return {'nullValue': value}

	elif isinstance(value, bytes):
		return {'bytesValue': b64encode(value).decode('utf-8')}

	elif isinstance(value, bool):
		return {'booleanValue': value}

	elif isinstance(value, int):
		return {'integerValue': value}

	elif isinstance(value, float):
		return {'doubleValue': value}

	elif isinstance(value, str):
		return {'stringValue': value}

	elif isinstance(value, dict):
		return {'mapValue': _to_datastore(value)}

	elif isinstance(value, datetime):
		return {'timestampValue': value.strftime("%Y-%m-%dT%H:%M:%S.%fZ")}

	elif isinstance(value, DatetimeWithNanoseconds):
		return {'timestampValue': value.rfc3339()}

	elif isinstance(value, GeoPoint):
		return {'geoPointValue': MessageToDict(value.to_protobuf())}
	else:

		raise TypeError("Cannot convert to a Firestore Value", value, "Invalid type", type(value))




#   Copyright (c) 2022 Asif Arman Rahman
#   Licensed under MIT (https://github.com/AsifArmanRahman/firebase/blob/main/LICENSE)

# --------------------------------------------------------------------------------------


import socket

from ._custom_sse_client import SSEClient


class ClosableSSEClient(SSEClient):

	def __init__(self, *args, **kwargs):
		self.should_connect = True
		super(ClosableSSEClient, self).__init__(*args, **kwargs)

	def _connect(self):
		if self.should_connect:
			super(ClosableSSEClient, self)._connect()
		else:
			raise StopIteration()

	def close(self):
		self.should_connect = False
		self.retry = 0

		self.resp.raw._fp.fp.raw._sock.shutdown(socket.SHUT_RDWR)
		self.resp.raw._fp.fp.raw._sock.close()
